from typing import List, Dict, Any

from rlgym.api import DoneCondition, AgentID, StateType
from rlgym.rocket_league.api import GameState


class CarOnGroundCondition(DoneCondition[AgentID, GameState]):
    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        pass

    def is_done(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[AgentID, bool]:
        return {agent: state.cars[agent].on_ground for agent in agents}



from typing import Dict, Any, Union

from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import OCTANE, DOMINUS, PLANK, BREAKOUT, HYBRID, MERC


class HitboxMutator(StateMutator[GameState]):
    """
    Sets the hitbox of all cars in the game to the specified hitbox type.
    """

    def __init__(self, hitbox_type: Union[int, str]):
        if isinstance(hitbox_type, str):
            hitbox_type = {
                "octane": OCTANE,
                "dominus": DOMINUS,
                "plank": PLANK,
                "breakout": BREAKOUT,
                "hybrid": HYBRID,
                "merc": MERC,
            }[hitbox_type.lower()]
        assert hitbox_type in (OCTANE, DOMINUS, PLANK, BREAKOUT, HYBRID, MERC), "Invalid hitbox"
        self.hitbox_type = hitbox_type

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        for car in state.cars.values():
            car.hitbox_type = self.hitbox_type



import warnings
from copy import deepcopy
from typing import Literal, Callable, Iterable

import numpy as np
from rlgym.rocket_league.api import GameState, Car
from rlgym.rocket_league.common_values import ORANGE_TEAM, BLUE_TEAM, TICKS_PER_SECOND, DOUBLEJUMP_MAX_DELAY, \
    FLIP_TORQUE_TIME, JUMP_MAX_TIME, CAR_MAX_SPEED, BOOST_LOCATIONS, SMALL_PAD_RECHARGE_SECONDS, \
    BIG_PAD_RECHARGE_SECONDS
from rlgym.rocket_league.math import quat_to_rot_mtx
from rlgym.rocket_league.sim import RocketSimEngine
from rlgym.rocket_league.state_mutators import FixedTeamSizeMutator, KickoffMutator

from rlgym_tools.rocket_league.math.ball import ball_hit_ground
from rlgym_tools.rocket_league.math.inverse_aerial_controls import aerial_inputs
from rlgym_tools.rocket_league.replays.parsed_replay import ParsedReplay
from rlgym_tools.rocket_league.replays.replay_frame import ReplayFrame
from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo

try:
    import numba
    import scipy  # Not used, but needed for numba

    optional_njit = numba.njit
except ImportError:
    warnings.warn("Numba/scipy not found, falling back to non-jitted functions")
    numba = None


    def optional_njit(f, *args, **kwargs):
        return f

_boost_locations = np.array(BOOST_LOCATIONS)


def replay_to_rlgym(
        replay: ParsedReplay,
        interpolation: Literal["none", "linear", "rocketsim"] = "rocketsim",
        predict_pyr: bool = True,
        calculate_error: bool = False,
        step_rounding: Callable[[float], int] = round,
        modify_action_fn: Callable[[Car, np.ndarray], np.ndarray] = None
) -> Iterable[ReplayFrame]:
    """
    Convert a parsed replay to a sequence of RLGym objects.

    :param replay: The parsed replay to convert
    :param interpolation: The interpolation mode to use. "none" uses the raw data, "linear" interpolates between frames,
                          "rocketsim" uses RocketSim to interpolate between frames.
    :param predict_pyr: Whether to predict pitch, yaw, roll from changes in angular velocity
                        (they're the only part of the action not provided by the replay)
    :param calculate_error: Whether to calculate the error between the data from the replay and the interpolated data.
    :param step_rounding: A function to round the number of ticks to step. Default is round.
    :param modify_action_fn: A function to modify the action before it is used. Takes car and action as arguments.
    :return: An iterable of ReplayFrame objects
    """
    if interpolation not in ("none", "linear", "rocketsim"):
        raise ValueError(f"Interpolation mode {interpolation} not recognized")
    rocketsim_interpolation = interpolation == "rocketsim"
    linear_interpolation = interpolation == "linear"

    players = {p["unique_id"]: p for p in replay.metadata["players"]}
    transition_engine = RocketSimEngine(rlbot_delay=False)
    count_orange = sum(p["is_orange"] is True for p in players.values())
    base_mutator = FixedTeamSizeMutator(blue_size=len(players) - count_orange, orange_size=count_orange)
    kickoff_mutator = KickoffMutator()

    hits = set()
    for hit in replay.analyzer.get("hits", []):
        f = hit["frame_number"]
        p = hit["player_unique_id"]
        hits.add((f, int(p)))

    average_tick_rate = replay.game_df["delta"].mean() * TICKS_PER_SECOND

    player_ids = sorted(int(k) for k in replay.player_dfs.keys())

    # Track scoreline
    blue = orange = 0
    final_goal_diff = sum(-1 if goal["is_orange"] else 1 for goal in replay.metadata["game"]["goals"])
    winning_team = None
    if final_goal_diff > 0:
        winning_team = BLUE_TEAM
    elif final_goal_diff < 0:
        winning_team = ORANGE_TEAM

    shared_info = {}
    gameplay_periods = replay.analyzer["gameplay_periods"]

    # For some reason transitioning from regulation to overtime is not always detected properly
    last_period = gameplay_periods[-1]
    start_frame = last_period["start_frame"]
    end_frame = last_period["end_frame"]
    overtime = replay.game_df.loc[start_frame:end_frame]["is_overtime"] == 1
    if overtime.nunique() > 1:
        # The last period is invalid, we need to split it up
        gameplay_periods = gameplay_periods[:-1]
        end_first = overtime.diff().idxmax()
        start_second = (replay.game_df["time"] > replay.game_df.loc[end_first, "time"] + 4).idxmax()
        first_hits = (replay.game_df.loc[start_frame:end_frame]["ball_has_been_hit"] == 1).diff() > 0
        first_hits = first_hits[first_hits].index
        goal_frame = end_frame if abs(replay.ball_df.loc[end_frame, "pos_y"]) > 5000 else None
        gameplay_periods.extend([
            {
                "start_frame": start_frame,
                "first_hit_frame": first_hits[0],
                "goal_frame": None,
                "end_frame": end_first - 1
            },
            {
                "start_frame": start_second,
                "first_hit_frame": first_hits[-1],
                "goal_frame": goal_frame,
                "end_frame": end_frame
            },
        ])

    for g, gameplay_period in enumerate(gameplay_periods):
        start_frame = gameplay_period["start_frame"]
        goal_frame = gameplay_period.get("goal_frame")
        end_frame = goal_frame or gameplay_period["end_frame"]

        # Prepare the starting state
        state = transition_engine.create_base_state()
        base_mutator.apply(state, shared_info)  # Base mutator creates the cars with default values

        # Rename agent ids and set cars to be the correct team
        new_cars = {}
        for uid, car in zip(replay.player_dfs, state.cars.values()):
            player = players[uid]
            # car.hitbox_type = OCTANE  # TODO: Get hitbox from replay
            car.team_num = ORANGE_TEAM if player["is_orange"] else BLUE_TEAM
            new_cars[int(uid)] = car
        state.cars = new_cars

        kickoff_mutator.apply(state, shared_info)
        if rocketsim_interpolation:
            actions = {uid: np.zeros((TICKS_PER_SECOND, 8)) for uid in player_ids}
            transition_engine.set_state(state, shared_info)
            # Step with no actions to let RocketSim calculate internal values
            state = transition_engine.step(actions, shared_info)

        # Prepare the segment dataframes (with some added columns)
        ball_df, game_df, player_dfs = _prepare_segment_dfs(replay, start_frame, end_frame,
                                                            linear_interpolation, predict_pyr)
        next_scoring_team = None
        if goal_frame is not None:
            # Get last index of ball_df, as it may have been shortened
            end_frame = goal_frame = ball_df.index[-1]
            next_scoring_team = BLUE_TEAM if ball_df["pos_y"].iloc[-1] > 0 else ORANGE_TEAM

        # Prepare tuples for faster iteration
        game_tuples = list(game_df.itertuples())
        ball_tuples = list(ball_df.itertuples())
        player_rows = []
        for pid in player_ids:
            pdf = player_dfs[pid]
            player_rows.append(list(pdf.itertuples()))

        assert ball_tuples[0].pos_x == ball_tuples[0].pos_y == 0

        # prev_actions = {uid: np.zeros(8) for uid in player_ids}

        # Iterate over the frames
        for i, game_row, ball_row, car_rows in zip(range(len(game_tuples)),
                                                   game_tuples,
                                                   ball_tuples,
                                                   zip(*player_rows)):
            frame = game_row.Index

            # Set the ball state. Unlike players, it updates every frame
            state.ball.position[:] = (ball_row.pos_x, ball_row.pos_y, ball_row.pos_z)
            state.ball.linear_velocity[:] = (ball_row.vel_x, ball_row.vel_y, ball_row.vel_z)
            state.ball.angular_velocity[:] = (ball_row.ang_vel_x, ball_row.ang_vel_y, ball_row.ang_vel_z)
            state.ball.quaternion = np.array((ball_row.quat_w, ball_row.quat_x,
                                              ball_row.quat_y, ball_row.quat_z))

            state.tick_count = game_row.time * TICKS_PER_SECOND
            state.goal_scored = frame == goal_frame

            actions = {}
            update_age = {}
            errors = {}

            for uid, player_row in zip(player_ids, car_rows):
                car = state.cars[uid]
                hit = (frame, uid) in hits

                if calculate_error:  # and i > 0 and not player_rows[player_ids.index(uid)][i - 1].is_repeat:
                    action, errs = _update_car_and_get_action(car, linear_interpolation, player_row, state,
                                                              calculate_error=True, hit=hit)
                    if (frame != start_frame  # Need to set the initial state first
                            and frame != goal_frame  # Goals can send players flying
                            and not car.is_demoed):  # Demoed cars just retain their physics from before the demo
                        errors[uid] = errs
                else:
                    action = _update_car_and_get_action(car, linear_interpolation, player_row, state, hit=hit)

                if modify_action_fn is not None:
                    action = modify_action_fn(car, action)

                actions[uid] = action
                update_age[uid] = player_row.update_age if not player_row.is_demoed else 0

            for uid, player_row in zip(player_ids, car_rows):
                # Detect demos so bump victims are correctly set
                if player_row.got_demoed:
                    closest = None
                    for uid2 in player_ids:
                        if uid == uid2:
                            continue
                        if state.cars[uid2].bump_victim_id == uid:
                            break  # Trust this if available
                        else:
                            dist = np.linalg.norm(state.cars[uid].physics.position - state.cars[uid2].physics.position)
                            if closest is None or dist < closest[1]:
                                closest = (uid2, dist)
                    else:
                        if closest[1] < 200 + 2 * CAR_MAX_SPEED * average_tick_rate / TICKS_PER_SECOND:
                            state.cars[closest[0]].bump_victim_id = uid

                # Detect boost grabs
                if player_row.boost_pickup > 0:
                    pad_dists = np.linalg.norm(_boost_locations - state.cars[uid].physics.position, axis=1)
                    closest_pad_idx = np.argmin(pad_dists)
                    if pad_dists[closest_pad_idx] < 200 and state.boost_pad_timers[closest_pad_idx] == 0:
                        if _boost_locations[closest_pad_idx, 1] > 71.5:
                            state.boost_pad_timers[closest_pad_idx] = BIG_PAD_RECHARGE_SECONDS
                        else:
                            state.boost_pad_timers[closest_pad_idx] = SMALL_PAD_RECHARGE_SECONDS

            if state.goal_scored:
                if state.scoring_team == BLUE_TEAM:
                    blue += 1
                elif state.scoring_team == ORANGE_TEAM:
                    orange += 1
            scoreboard = ScoreboardInfo(
                game_timer_seconds=game_row.scoreboard_timer,
                kickoff_timer_seconds=game_row.kickoff_timer,
                blue_score=blue,
                orange_score=orange,
                go_to_kickoff=frame == end_frame and g < len(gameplay_periods) - 1,
                is_over=(frame == end_frame
                         and g == len(gameplay_periods) - 1
                         and blue != orange
                         and not (1 < game_row.scoreboard_timer <= 300)
                         and (state.goal_scored
                              or ball_hit_ground(2 * average_tick_rate, state.ball, pre=True)
                              or ball_hit_ground(game_row.delta * TICKS_PER_SECOND, state.ball, pre=False))),
            )

            episode_seconds_remaining = game_row.episode_seconds_remaining

            res = ReplayFrame(
                state=state,
                actions=actions,
                update_age=update_age,
                scoreboard=scoreboard,
                episode_seconds_remaining=episode_seconds_remaining,
                next_scoring_team=next_scoring_team,
                winning_team=winning_team,
            )
            if calculate_error:
                yield res, errors
            else:
                yield res

            if frame == end_frame:
                break

            ticks = step_rounding(game_tuples[i + 1].time * TICKS_PER_SECOND - state.tick_count)
            for uid, action in actions.items():
                repeated_action = action.reshape(1, -1).repeat(ticks, axis=0)
                # if ticks > 1:
                #     repeated_action[:1, :] = prev_actions[uid]  # Simulate rlbot_delay
                # prev_actions[uid] = action  # Keep the unrepeated action for the next frame
                actions[uid] = repeated_action

            if rocketsim_interpolation:
                transition_engine.set_state(state, {})
                state = transition_engine.step(actions, {})
            else:
                state = deepcopy(state)


def _prepare_segment_dfs(replay, start_frame, end_frame, linear_interpolation, predict_pyr):
    ball_df = replay.ball_df.loc[start_frame:end_frame].astype(float)
    # Find the first frame where the ball is sleeping and not at the origin, that's when the goal actually happens       
    actual_goal = ball_df.index[(ball_df["is_sleeping"] == 1) & (ball_df["pos_x"] != 0) & (ball_df["pos_y"] != 0)]
    if len(actual_goal) > 0 and actual_goal[0] < end_frame:
        end_frame = actual_goal[0]
        ball_df = ball_df.loc[:end_frame]
    game_df = replay.game_df.loc[start_frame:end_frame].astype(float)

    # Fill missing ball values
    ball_df["quat_w"] = ball_df["quat_w"].fillna(1.0)
    ball_df = ball_df.ffill().fillna(0.)

    # Players
    player_dfs = {}
    for uid, pdf in replay.player_dfs.items():
        pdf = pdf.loc[start_frame:end_frame].astype(float)
        physics_cols = ([f"{col}_{axis}" for col in ("pos", "vel", "ang_vel") for axis in "xyz"] +
                        [f"quat_{axis}" for axis in "wxyz"])

        # Physics info is repeated 1-3 times, so we need to not use those values
        is_repeat = (pdf[physics_cols].diff() == 0).all(axis=1)

        # If something repeats for 4 or more frames, assume it's not a real repeat, just the player standing still
        is_repeat &= (is_repeat.rolling(4).sum() < 4)

        # Physics info is nan while the player is demoed
        is_demoed = pdf[["is_sleeping"] + physics_cols].isna().any(axis=1)

        # Interpolate, smooth, and shift (essentially shift by -0.5)
        pdf[["throttle", "steer"]] = pdf[["throttle", "steer"]].ffill().fillna(255 / 2)
        pdf.loc[is_repeat, ["throttle", "steer"]] = np.nan
        pdf[["throttle", "steer"]] = (pdf[["throttle", "steer"]].interpolate()
                                      .rolling(2).mean().shift(-1, fill_value=255 / 2))

        # Remove nans
        pdf = pdf.ffill().fillna(0.)
        pdf["is_repeat"] = is_repeat
        pdf["is_demoed"] = is_demoed.astype(float)
        pdf["got_demoed"] = pdf["is_demoed"].diff() > 0
        pdf["respawned"] = pdf["is_demoed"].diff() < 0

        # Demos while flipping make dodge_is_active stay true until they respawn and flip again
        # For these cases we need to find the next frame where dodge torque changes, which is when they flip again,
        # and fill the dodge_is_active column with 0 until then
        dodge_active_on_respawn = (pdf["dodge_is_active"] == 1) & pdf["respawned"]
        if dodge_active_on_respawn.any():
            torque_change = (pdf["dodge_torque_x"].diff() != 0) | (pdf["dodge_torque_y"].diff() != 0)
            start_idx = dodge_active_on_respawn[dodge_active_on_respawn].index
            end_idx = torque_change[torque_change].index
            for i in range(len(start_idx)):
                for j in range(len(end_idx)):
                    if end_idx[j] > start_idx[i]:
                        pdf.loc[start_idx[i]:end_idx[j] - 1, "dodge_is_active"] = 0
                        break

        # We need to make sure we do some actions 1 frame before so they're active when they're supposed to be
        # We treat jump and dodge differently on first frame since the requirements are different
        pdf["jumped"] = (pdf["jump_is_active"].diff() > 0).shift(-1, fill_value=False)
        pdf["dodged"] = (pdf["dodge_is_active"].diff() > 0).shift(-1, fill_value=False)
        pdf[["dodge_torque_x", "dodge_torque_y"]] = pdf[["dodge_torque_x", "dodge_torque_y"]].shift(-1, fill_value=0.)
        pdf["double_jumped"] = (pdf["double_jump_is_active"].diff() > 0).shift(-1, fill_value=False)
        boosted = (pdf["boost_is_active"].diff() > 0).shift(-1, fill_value=False)
        pdf.loc[boosted, "boost_is_active"] = 1
        flipped_car = (pdf["flip_car_is_active"].diff() > 0).shift(-1, fill_value=False)
        pdf.loc[flipped_car, "flip_car_is_active"] = 1

        # Make sure jump isn't active right before a dodge or double jump
        m = (((pdf["dodge_is_active"].shift(-2, fill_value=0) == 1)
              | (pdf["double_jump_is_active"].shift(-2, fill_value=0) == 1))
             & (pdf["jump_is_active"] == 1))
        pdf.loc[m, "jump_is_active"] = 0

        times = game_df["time"].copy()
        times[is_repeat] = np.nan
        times = times.ffill().fillna(0.)
        pdf["update_age"] = game_df["time"] - times

        pyr_cols = ["pitch", "yaw", "roll"]
        pdf[pyr_cols] = np.nan
        if predict_pyr:
            pyrs = pyr_from_dataframe(game_df, pdf)
            pdf.loc[~is_repeat, pyr_cols] = pyrs
        # ffill because pyr calculation makes a smooth transition from one angvel to the next
        pdf[pyr_cols] = pdf[pyr_cols].ffill().fillna(0.)

        # Player physics info is repeated 1-3 times, so we need to remove those
        if linear_interpolation:
            # Set interpolate cols to NaN if they are repeated
            pdf.loc[is_repeat, physics_cols] = np.nan
            pdf = pdf.interpolate()  # Note that this assumes equal time steps

        player_dfs[int(uid)] = pdf
    game_df["episode_seconds_remaining"] = game_df["time"].iloc[-1] - game_df["time"]

    # `seconds_remaining` seems to be the ceil of the true game timer
    # We recreate the true timer here
    scoreboard_timer = game_df["seconds_remaining"].copy()
    ball_hit = game_df["ball_has_been_hit"] == 1  # Where the timer is running

    # Synchronize by finding first change in the timer
    decreases = scoreboard_timer[ball_hit].diff() < 0
    if decreases.any():
        first_change = decreases.idxmax()
    else:
        first_change = decreases.index[0]
    delta = game_df["time"][ball_hit] - game_df["time"][ball_hit].loc[first_change]
    scoreboard_timer[ball_hit] = scoreboard_timer.loc[first_change] - delta
    scoreboard_timer[~ball_hit] = np.nan
    scoreboard_timer = scoreboard_timer.ffill().bfill()

    # Handle overtime and negative values
    if "is_overtime" in game_df:
        scoreboard_timer[game_df["is_overtime"] == 1] = np.inf
    scoreboard_timer[scoreboard_timer < 0] = 0

    game_df["scoreboard_timer"] = scoreboard_timer

    # Now kickoff timer
    game_df["kickoff_timer"] = 0.0
    try:
        delta = game_df["time"][~ball_hit] - game_df["time"][~ball_hit].iloc[0]
        kickoff_timer = (5.0 - delta).clip(0, 5)
    except IndexError:
        kickoff_timer = 0.0
    game_df.loc[~ball_hit, "kickoff_timer"] = kickoff_timer

    return ball_df, game_df, player_dfs


def _update_car_and_get_action(car: Car, linear_interpolation: bool, player_row, state: GameState,
                               calculate_error=False, hit: bool = False):
    pre_car = deepcopy(car) if calculate_error else None
    replay_update = not player_row.is_repeat
    if linear_interpolation or replay_update:
        true_pos = (player_row.pos_x, player_row.pos_y, player_row.pos_z)
        true_vel = (player_row.vel_x, player_row.vel_y, player_row.vel_z)
        true_ang_vel = (player_row.ang_vel_x, player_row.ang_vel_y, player_row.ang_vel_z)
        true_quat = (player_row.quat_w, player_row.quat_x, player_row.quat_y, player_row.quat_z)

        car.physics.position[:] = true_pos
        car.physics.linear_velocity[:] = true_vel
        car.physics.angular_velocity[:] = true_ang_vel
        car.physics.quaternion = np.array(true_quat)  # Uses property setter

    if player_row.got_demoed and not car.is_demoed:
        car.demo_respawn_timer = 3
    elif player_row.is_demoed and not car.is_demoed:
        car.demo_respawn_timer = 1 / TICKS_PER_SECOND
    elif not player_row.is_demoed and car.is_demoed:
        car.demo_respawn_timer = 0

    car.boost_amount = player_row.boost_amount
    throttle = 2 * player_row.throttle / 255 - 1
    steer = 2 * player_row.steer / 255 - 1
    if abs(throttle) < 0.01:
        throttle = 0
    if abs(steer) < 0.01:
        steer = 0
    pitch = player_row.pitch
    yaw = player_row.yaw
    roll = player_row.roll
    jump = 0

    if player_row.dodged or player_row.double_jumped:
        # Make sure the car is in a valid state for dodging/double jumping
        car.has_flipped = False
        if car.on_ground:
            car.on_ground = False
        car.is_holding_jump = False
        car.has_double_jumped = False
        if car.air_time_since_jump >= DOUBLEJUMP_MAX_DELAY:
            car.air_time_since_jump = DOUBLEJUMP_MAX_DELAY - 1 / TICKS_PER_SECOND
        assert car.can_flip

    if player_row.jumped:
        if not car.on_ground:
            car.on_ground = True
        jump = 1
    elif player_row.jump_is_active:
        # assert not car.on_ground
        # Player is holding jump
        if car.on_ground:
            car.on_ground = False
        car.air_time_since_jump = 0
        car.is_jumping = True
        car.is_holding_jump = True
        car.has_jumped = True
        car.has_flipped = False
        car.has_double_jumped = False
        if car.jump_time > JUMP_MAX_TIME:
            car.jump_time = JUMP_MAX_TIME - 1 / TICKS_PER_SECOND
        jump = 1
    elif player_row.dodged:
        new_pitch = -player_row.dodge_torque_y
        new_roll = -player_row.dodge_torque_x
        mx = max(abs(new_pitch), abs(new_roll))
        if mx > 0:
            # Project to unit square because why not
            pitch = new_pitch / mx
            yaw = 0
            roll = new_roll / mx
        else:
            # Stall, happens when roll+yaw=0 and abs(roll)+abs(yaw)>deadzone
            roll = 1 if roll > 0 else -1
            yaw = -roll
            pitch = 0
        jump = 1
    elif player_row.dodge_is_active:
        actual_torque = np.array([player_row.dodge_torque_x, player_row.dodge_torque_y, 0])
        actual_torque = actual_torque / (np.linalg.norm(actual_torque) or 1)

        if car.on_ground:
            car.on_ground = False
        # car.is_flipping = True
        car.flip_torque = actual_torque
        car.has_flipped = True
        car.has_double_jumped = False
        if car.flip_time <= 0:
            car.flip_time = 1 / TICKS_PER_SECOND
        elif car.flip_time >= FLIP_TORQUE_TIME:
            car.flip_time = FLIP_TORQUE_TIME - 1 / TICKS_PER_SECOND
        assert not car.has_flip
        # Pitch/yaw/roll is handled by inverse aerial control function already,
        # it knows about the flip and detects the cancel automatically
        # pitch = yaw = roll = 0
    elif player_row.double_jumped:
        mag = abs(pitch) + abs(yaw) + abs(roll)
        if mag >= state.config.dodge_deadzone:
            # Would not have been a double jump, but a dodge. Correct it
            # {m>d, l*m<d} -> c<d/m for d>0 and m>d
            # New magnitude will be 0.49 with default deadzone
            limiter = 0.98 * state.config.dodge_deadzone / mag
            pitch = pitch * limiter
            roll = roll * limiter
            yaw = yaw * limiter
        jump = 1
        assert car.has_flip
    elif player_row.double_jump_is_active:
        if car.on_ground:
            car.on_ground = False
        car.has_double_jumped = True
        car.has_flipped = False
        assert not car.has_flip
    elif player_row.flip_car_is_active and not car.is_autoflipping:
        # I think flip_car is autoflip?
        if car.on_ground:
            car.on_ground = False
        jump = 1

    if car.on_ground:
        car.air_time_since_jump = 0
        # car.jump_time = 0  # Not reset for whatever reason, might be in the future
        car.flip_time = 0
        car.has_flipped = False
        car.has_jumped = False
        car.has_double_jumped = False

    if hit and car.ball_touches == 0:
        car.ball_touches = 1
    boost = player_row.boost_is_active
    handbrake = player_row.handbrake
    action = np.array([throttle, steer, pitch, yaw, roll, jump, boost, handbrake],
                      dtype=np.float32)

    if calculate_error:
        errors = {}
        if pre_car is not None:
            if replay_update and not (player_row.respawned or (player_row.is_demoed and not player_row.got_demoed)):
                # During demos the physics state isn't really valid, and they can respawn on either side
                for slot in car.physics.__slots__:  # position, linear_velocity, etc.
                    if slot.startswith("_"):
                        continue
                    old_val = getattr(pre_car.physics, slot)
                    new_val = getattr(car.physics, slot)
                    errors[slot] = np.linalg.norm(new_val - old_val)
                # Quaternions have two valid representations, so we need to check both
                new_val = car.physics.quaternion
                old_val = pre_car.physics.quaternion
                errors["quaternion"] = 1 - np.dot(new_val, old_val) ** 2

            for slot in car.__slots__:
                if slot == "physics" or slot.startswith("_"):
                    continue
                new_val = getattr(car, slot)
                old_val = getattr(pre_car, slot)
                if isinstance(new_val, np.ndarray):
                    if slot == "flip_torque" and not player_row.dodge_is_active:
                        # We shifted the torque so we could use it to get the correct action,
                        # but the state won't update until next frame
                        continue
                    errors[slot] = np.linalg.norm(new_val - old_val)
                elif isinstance(new_val, (float, int, bool)):
                    errors[slot] = abs(new_val - old_val)
                else:
                    errors[slot] = new_val != old_val
        return action, errors

    return action


_numba_quat_2_rot = optional_njit(quat_to_rot_mtx)
_numba_aerial_inputs = optional_njit(aerial_inputs)


@optional_njit
def _multi_aerial_inputs(quats, ang_vels, times, is_flipping):
    pyrs = np.zeros((len(quats), 3), dtype=np.float32)
    next_rot = _numba_quat_2_rot(quats[0])
    for i in range(len(quats) - 1):
        rot = next_rot
        next_rot = _numba_quat_2_rot(quats[i + 1])
        dt = times[i + 1] - times[i]
        pyrs[i, :] = _numba_aerial_inputs(ang_vels[i], ang_vels[i + 1], rot, next_rot, dt, is_flipping[i])
    return pyrs


def pyr_from_dataframe(game_df, player_df):
    is_repeated = player_df["is_repeat"].values
    times = game_df["time"].values.astype(np.float32)
    ang_vels = player_df[['ang_vel_x', 'ang_vel_y', 'ang_vel_z']].values.astype(np.float32)
    quats = player_df[['quat_w', 'quat_x', 'quat_y', 'quat_z']].values.astype(np.float32)
    is_flipping = player_df["dodge_is_active"].values.astype(bool)

    quats = quats[~is_repeated]
    ang_vels = ang_vels[~is_repeated]
    times = times[~is_repeated]
    is_flipping = is_flipping[~is_repeated]

    pyrs = _multi_aerial_inputs(quats, ang_vels, times, is_flipping)

    return pyrs



import os.path
import random
from typing import Dict, Any, List, Literal, Tuple, Union, Optional

import numpy as np
from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.misc.serialize import GS_CARS, GS_CAR_LENGTH, RF_AGENT_IDS_START, serialize_replay_frame, \
    deserialize_replay_frame, RF_ACTION_SIZE
from rlgym_tools.rocket_league.replays.convert import replay_to_rlgym
from rlgym_tools.rocket_league.replays.parsed_replay import ParsedReplay
from rlgym_tools.rocket_league.replays.replay_frame import ReplayFrame


class ReplayMutator(StateMutator[GameState]):
    def __init__(
            self,
            replay_frames: Union[str, np.ndarray],
            probabilities: Union[None, str, np.ndarray] = None,
    ):
        """
        A state mutator that randomly selects a state from a replay file and applies it to the current state.

        :param replay_frames: The replay file to sample states from. Can be a .npz file, a .dat file, or a numpy array.
        :param probabilities: The probabilities of selecting each state in the replay file. Defaults to uniform.
        """

        if isinstance(replay_frames, str):
            if replay_frames.endswith(".npz"):
                with np.load(replay_frames) as data:
                    replay_frames = data["replay_frames"]
                    if probabilities is None:
                        probabilities = data.get("probabilities")
            elif replay_frames.endswith(".dat"):
                replay_frames = np.memmap(replay_frames, dtype='float32', mode='r')
            else:
                raise ValueError("Invalid replay file")

        if isinstance(probabilities, str):
            probabilities = np.load(probabilities)  # Assume .npy file

        self.replay_frames = replay_frames
        self.probabilities = self.assign_probabilities() if probabilities is None else probabilities

    def assign_probabilities(self):
        """
        Assigns probabilities to each state in the replay file.
        Can be overridden to provide custom probabilities based on the states.

        :return: A list of probabilities for each state.
        """
        return None  # Uniform distribution

    def __getitem__(self, item) -> ReplayFrame:
        return deserialize_replay_frame(self.replay_frames[item])

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        idx = np.random.choice(len(self.replay_frames), p=self.probabilities)
        replay_frame = self[idx]
        new_state = replay_frame.state
        state.tick_count = round(new_state.tick_count)
        state.goal_scored = new_state.goal_scored
        state.config = new_state.config
        state.cars = new_state.cars
        state.ball = new_state.ball
        state.boost_pad_timers = new_state.boost_pad_timers

        shared_info["replay_frame"] = replay_frame  # In case anyone needs more than state and scoreboard
        shared_info["scoreboard"] = replay_frame.scoreboard

    @staticmethod
    def make_file(replay_files: List[str],
                  output_path: Optional[str],
                  do_memory_map: bool = False,
                  frame_skip: int = 30,
                  interpolation: Literal["none", "linear", "rocketsim"] = "rocketsim",
                  carball_path=None,
                  max_num_players=6) -> np.ndarray:
        size = len(replay_files) * 5 * 60 * 30 // frame_skip  # Initial guess of frame count

        max_state_size = GS_CARS.start + max_num_players * GS_CAR_LENGTH
        max_replay_frame_size = (RF_AGENT_IDS_START
                                 + 2 * max_num_players  # agent_id, update_age
                                 + max_num_players * RF_ACTION_SIZE  # actions
                                 + max_state_size)

        if do_memory_map:
            replay_frames = np.memmap(os.path.join(output_path, "replay_frames.dat"),
                                      dtype='float32', mode='w+', shape=(size, max_replay_frame_size))
        else:
            replay_frames = np.zeros((size, max_replay_frame_size), dtype=np.float32)

        i = 0
        for k, replay_file in enumerate(replay_files):
            parsed_replay = ParsedReplay.load(replay_file, carball_path)
            frame = random.randrange(0, frame_skip)  # Randomize starting frame to increase diversity
            for replay_frame in replay_to_rlgym(parsed_replay, interpolation, predict_pyr=True, calculate_error=False):
                if frame % frame_skip == 0:
                    serialized_frame = serialize_replay_frame(replay_frame)
                    # if (frame // frame_skip) % frame_skip == 0:
                    #     d1 = deserialize_replay_frame(serialized_frame)
                    #     s2 = np.zeros(max_replay_frame_size, dtype=np.float32)
                    #     s2[:len(serialized_frame)] = serialized_frame
                    #     d2 = deserialize_replay_frame(s2)
                    #     debug = True

                    if i >= size:
                        # Make a new estimate of the size of the replay files and expand the arrays
                        frames_per_replay = (i + 1) / (k or 1)
                        new_size = (len(replay_files) + 1) * frames_per_replay  # +1 so we don't have to resize a lot
                        new_size = round(new_size)
                        # Shouldn't need refcheck, just makes it fail in debug mode
                        replay_frames.resize((new_size, max_replay_frame_size), refcheck=False)

                    replay_frames[i, :len(serialized_frame)] = serialized_frame

                    i += 1
                frame += 1

        replay_frames.resize((i, max_replay_frame_size), refcheck=False)

        if not do_memory_map and output_path is not None:
            np.savez_compressed(output_path, replay_frames=replay_frames)

        return replay_frames



from typing import Any, Dict, Sequence, Tuple

import numpy as np
from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState


class WeightedSampleMutator(StateMutator[GameState]):
    def __init__(self, mutators: Sequence[StateMutator], weights: Sequence[float]):
        assert len(mutators) == len(weights)
        self.mutators = mutators
        weights = np.array(weights)
        self.probs = weights / weights.sum()

    @staticmethod
    def from_zipped(*mutator_weights: Tuple[StateMutator, float]):
        mutators, weights = zip(*mutator_weights)
        return WeightedSampleMutator(mutators, weights)

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        idx = np.random.choice(len(self.mutators), p=self.probs)
        mutator = self.mutators[idx]
        mutator.apply(state, shared_info)



from typing import List, Dict, Any

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState


class FlipResetReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, obtain_flip_weight: float = 1.0, hit_ball_weight: float = 1.0):
        self.obtain_flip_weight = obtain_flip_weight
        self.hit_ball_weight = hit_ball_weight

        self.prev_state = None
        self.has_reset = None
        self.has_flipped = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prev_state = initial_state
        self.has_reset = set()
        self.has_flipped = set()

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = {k: 0 for k in agents}
        for agent in agents:
            car = state.cars[agent]
            if car.on_ground:
                self.has_reset.discard(agent)
                self.has_flipped.discard(agent)
            elif car.can_flip and not self.prev_state.cars[agent].can_flip:
                down = -car.physics.up
                car_ball = state.ball.position - car.physics.position
                car_ball /= np.linalg.norm(car_ball)
                cossim_down_ball = np.dot(down, car_ball)
                if cossim_down_ball > 0.5 ** 0.5:  # 45 degrees
                    self.has_reset.add(agent)
                    rewards[agent] = self.obtain_flip_weight
            elif car.is_flipping and agent in self.has_reset:
                self.has_reset.remove(agent)
                self.has_flipped.add(agent)
            if car.ball_touches > 0 and agent in self.has_flipped:
                self.has_flipped.remove(agent)
                rewards[agent] = self.hit_ball_weight
        self.prev_state = state
        return rewards



from typing import Any, Dict, List

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import CAR_MAX_SPEED


class WavedashReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, scale_by_acceleration: bool = True):
        self.scale_by_acceleration = scale_by_acceleration
        self.prev_state = None
        self.prev_acceleration = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prev_state = initial_state
        self.prev_acceleration = {agent: 0 for agent in agents}

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = {}
        for agent in agents:
            car = state.cars[agent]
            prev_car = self.prev_state.cars[agent]

            wavedash = ((car.on_ground and not prev_car.on_ground)
                        and (car.is_flipping or prev_car.is_flipping))

            if self.scale_by_acceleration:
                if car.is_flipping and not prev_car.is_flipping:
                    acc = np.linalg.norm(car.physics.linear_velocity - prev_car.physics.linear_velocity)
                    self.prev_acceleration[agent] = acc
                if wavedash:
                    acc = self.prev_acceleration[agent]
                    rewards[agent] = acc / CAR_MAX_SPEED
                    self.prev_acceleration[agent] = 0
                else:
                    rewards[agent] = 0
                    if not car.is_flipping:
                        self.prev_acceleration[agent] = 0
            else:
                if wavedash:
                    rewards[agent] = 1
                else:
                    rewards[agent] = 0
        self.prev_state = state

        return rewards



from typing import Dict, Any

import numpy as np
from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import SIDE_WALL_X, BACK_WALL_Y, CEILING_Z
from rlgym.rocket_league.math import rand_vec3, rand_uvec3

from rlgym_tools.rocket_league.reward_functions.aerial_distance_reward import RAMP_HEIGHT


class RandomPhysicsMutator(StateMutator[GameState]):
    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        padding = 100  # Ball radius and car hitbox with biggest radius are both below this
        for po in [state.ball] + [car.physics for car in state.cars.values()]:
            while True:
                new_pos = np.random.uniform([-SIDE_WALL_X + padding, -BACK_WALL_Y + padding, 0 + padding],
                                            [SIDE_WALL_X - padding, BACK_WALL_Y - padding, CEILING_Z - padding])

                # Some checks to make sure we don't place it outside the field
                if abs(new_pos[0]) + abs(new_pos) >= 8064 - padding:
                    continue
                close_to_wall = (abs(new_pos[0]) >= SIDE_WALL_X - RAMP_HEIGHT
                                 or abs(new_pos[1]) >= BACK_WALL_Y - RAMP_HEIGHT
                                 or abs(new_pos[0]) + abs(new_pos[1]) >= 8064 - RAMP_HEIGHT)
                close_to_floor_or_ceiling = (new_pos[2] <= RAMP_HEIGHT
                                             or new_pos[2] >= CEILING_Z - RAMP_HEIGHT)
                if close_to_wall and close_to_floor_or_ceiling:
                    continue
                break
            po.position = new_pos
            po.linear_velocity = rand_vec3(2300)
            po.angular_velocity = rand_vec3(5)
            fw = rand_uvec3()
            up = rand_uvec3()
            right = np.cross(fw, up)
            po.rotation_mtx = np.stack([fw, right, up])



from rlgym_tools.rocket_league.v1.objects.game_state import V1GameState
from rlgym_tools.rocket_league.v1.objects.physics_object import V1PhysicsObject
from rlgym_tools.rocket_league.v1.objects.player_data import V1PlayerData



import json
import socket
from typing import Dict, Any

import numpy as np
from rlgym.api import Renderer
from rlgym.rocket_league.api import GameState, Car

DEFAULT_UDP_IP = "127.0.0.1"
DEFAULT_UDP_PORT = 9273  # Default RocketSimVis port

BUTTON_NAMES = ("throttle", "steer", "pitch", "yaw", "roll", "jump", "boost", "handbrake")


class RocketSimVisRenderer(Renderer[GameState]):
    """
    A renderer that sends game state information to RocketSimVis.

    This is just the client side, you need to run RocketSimVis to see the visualization.
    Code is here: https://github.com/ZealanL/RocketSimVis
    """
    def __init__(self, udp_ip=DEFAULT_UDP_IP, udp_port=DEFAULT_UDP_PORT):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP
        self.udp_ip = udp_ip
        self.udp_port = udp_port

    @staticmethod
    def write_physobj(physobj):
        j = {
            'pos': physobj.position.tolist(),
            'forward': physobj.forward.tolist(),
            'up': physobj.up.tolist(),
            'vel': physobj.linear_velocity.tolist(),
            'ang_vel': physobj.angular_velocity.tolist()
        }

        return j

    @staticmethod
    def write_car(car: Car, controls=None):
        j = {
            'team_num': int(car.team_num),
            'phys': RocketSimVisRenderer.write_physobj(car.physics),
            'boost_amount': car.boost_amount,
            'on_ground': bool(car.on_ground),
            "has_flipped_or_double_jumped": bool(car.has_flipped or car.has_double_jumped),
            'is_demoed': bool(car.is_demoed),
            'has_flip': bool(car.can_flip)
        }

        if controls is not None:
            if isinstance(controls, np.ndarray):
                controls = {
                    k: float(v)
                    for k, v in zip(BUTTON_NAMES, controls)
                }
            j['controls'] = controls

        return j

    def render(self, state: GameState, shared_info: Dict[str, Any]) -> Any:
        if "controls" in shared_info:
            controls = shared_info["controls"]
        else:
            controls = {}
        j = {
            'ball_phys': self.write_physobj(state.ball),
            'cars': [
                self.write_car(car, controls.get(agent_id))
                for agent_id, car in state.cars.items()
            ],
            'boost_pad_states': (state.boost_pad_timers <= 0).tolist()
        }

        self.sock.sendto(json.dumps(j).encode('utf-8'), (self.udp_ip, self.udp_port))

    def close(self):
        pass



from typing import List, Dict, Any

from rlgym.api import DoneCondition, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BLUE_TEAM, BALL_RADIUS
from rlgym.rocket_league.common_values import TICKS_PER_SECOND

from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


class GameCondition(DoneCondition[AgentID, GameState]):
    """
    Simulates a Rocket League game, ending when:
    - A goal is scored
    - The ball hits the ground at 0 seconds
    - The overtime exceeds the maximum allowed time
    - The scoreline is presumed to be insurmountable, leading to a forfeit
    """

    def __init__(self, seconds_per_goal_forfeit=None, max_overtime_seconds=None):
        self.seconds_per_goal_forfeit = seconds_per_goal_forfeit
        self.max_overtime_seconds = max_overtime_seconds
        self.overtime_duration = 0
        self.last_ticks = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.overtime_duration = 0
        self.last_ticks = initial_state.tick_count

    def is_done(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[AgentID, bool]:
        scoreboard: ScoreboardInfo = shared_info["scoreboard"]

        done = False
        if scoreboard.go_to_kickoff or scoreboard.is_over:
            done = True
        elif scoreboard.is_overtime:
            self.overtime_duration += (state.tick_count - self.last_ticks) / TICKS_PER_SECOND
            if self.max_overtime_seconds is not None and self.overtime_duration >= self.max_overtime_seconds:
                scoreboard.is_over = True
                done = True
        else:
            if self.seconds_per_goal_forfeit is not None:
                goal_diff = abs(scoreboard.blue_score - scoreboard.orange_score)
                if goal_diff >= 3:
                    seconds_per_goal = scoreboard.game_timer_seconds / goal_diff
                    if seconds_per_goal < self.seconds_per_goal_forfeit:  # Forfeit if it's not realistic to catch up
                        scoreboard.is_over = True
                        done = True

        self.last_ticks = state.tick_count

        return {agent: done for agent in agents}



from typing import Any, Dict, List

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import CAR_MAX_SPEED, BALL_MAX_SPEED, SIDE_WALL_X, BACK_WALL_Y, CEILING_Z, \
    BALL_RADIUS


class VelocityPlayerToBallReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, include_negative_values: bool = True, use_trajectory_comparison: bool = False,
                 use_dot_quotient: bool = False):
        self.include_negative_values = include_negative_values
        self.use_trajectory_comparison = use_trajectory_comparison
        self.use_dot_quotient = use_dot_quotient

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        pass

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        return {agent: self._get_reward(agent, state) for agent in agents}

    def _get_reward(self, agent: AgentID, state: GameState):
        ball = state.ball
        car = state.cars[agent].physics
        if self.use_trajectory_comparison:
            curr_dist, min_dist, t = trajectory_comparison(car.position, car.linear_velocity,
                                                           ball.position, ball.linear_velocity)
            vel = (curr_dist - min_dist) / t if t != 0 else 0
            norm_vel = vel / (CAR_MAX_SPEED + BALL_MAX_SPEED)
            if abs(norm_vel) > 1:  # In case of floating point errors with small t
                norm_vel = np.sign(norm_vel)
        elif self.use_dot_quotient:
            car_to_ball = ball.position - car.position
            car_to_ball = car_to_ball / np.linalg.norm(car_to_ball)

            # Vector version of v=d/t <=> t=d/v <=> 1/t=v/d which becomes v . d / |d|^2
            # Max value should be max_speed / ball_radius = 2300 / 92.75 = 24.8
            vd = np.dot(car_to_ball, car.linear_velocity)
            dd = np.dot(car_to_ball, ball.linear_velocity)
            inv_time = vd / dd if dd != 0 else 0
            norm_vel = inv_time / (CAR_MAX_SPEED / BALL_RADIUS)
        else:
            car_to_ball = ball.position - car.position
            car_to_ball = car_to_ball / np.linalg.norm(car_to_ball)

            vel = np.dot(car_to_ball, car.linear_velocity)
            norm_vel = vel / CAR_MAX_SPEED
        if self.include_negative_values:
            return norm_vel
        return max(0, norm_vel)


def trajectory_comparison(pos1, vel1, pos2, vel2, check_bounds=True):
    """
    Calculate the closest point between two trajectories, defined as the lines:
      pos1 + t * vel1
      pos2 + t * vel2
    """
    # First, find max time based on field bounds
    if check_bounds:
        max_time = np.inf
        for pos, vel in (pos1, vel1), (pos2, vel2):
            bounds = np.array([[-SIDE_WALL_X, -BACK_WALL_Y, 0],
                               [SIDE_WALL_X, BACK_WALL_Y, CEILING_Z]])
            times = (bounds - pos) / (vel + (vel == 0))
            times = times[times > 0]
            t = np.min(times)
            max_time = min(max_time, t)

    # The distance between the two rays is `||pos1 + t * vel1 - pos2 - t * vel2||`
    # This is equivalent to `||(pos1 - pos2) + t * (vel1 - vel2)||`
    pos_diff = pos1 - pos2
    vel_diff = vel1 - vel2

    # The minimum distance is achieved when the derivative of the distance is 0.
    # E.g. `d/dt * sqrt((p_x+t*v_x)^2+(p_y+t*v_y)^2+(p_z+t*v_z)^2)=0`
    # This is equivalent to
    #    `d/dt * (p_x+t*v_x)^2+(p_y+t*v_y)^2+(p_z+t*v_z)^2=0`
    # => `2*(p_x+t*v_x)*v_x+2*(p_y+t*v_y)*v_y+2*(p_z+t*v_z)*v_z=0`
    # => `p_x*v_x+p_y*v_y+p_z*v_z+t*(v_x^2+v_y^2+v_z^2)=0`
    # => `t=-(p_x*v_x+p_y*v_y+p_z*v_z)/(v_x^2+v_y^2+v_z^2)`
    denom = np.dot(vel_diff, vel_diff)
    if denom == 0:
        t = 0
    else:
        t = -np.dot(pos_diff, vel_diff) / denom

    if t > max_time:
        t = max_time

    # The minimum distance is then the distance at this time.
    curr_dist = np.linalg.norm(pos_diff)
    min_dist = np.linalg.norm(pos_diff + t * vel_diff)

    return curr_dist, min_dist, t



from typing import Dict, Any

from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState


class ConfigMutator(StateMutator[GameState]):
    def __init__(self, gravity: float = 1.0, boost_consumption: float = 1.0, dodge_deadzone: float = 0.5):
        self.gravity = gravity
        self.boost_consumption = boost_consumption
        self.dodge_deadzone = dodge_deadzone

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        state.config.gravity = self.gravity
        state.config.boost_consumption = self.boost_consumption
        state.config.dodge_deadzone = self.dodge_deadzone



import math
from dataclasses import dataclass, asdict, replace
from typing import List, Dict, Any

from rlgym.api import SharedInfoProvider, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import TICKS_PER_SECOND, BLUE_TEAM

from rlgym_tools.rocket_league.math.ball import ball_hit_ground


@dataclass(slots=True)
class ScoreboardInfo:
    game_timer_seconds: float
    kickoff_timer_seconds: float
    blue_score: int
    orange_score: int
    go_to_kickoff: bool
    is_over: bool

    @property
    def is_overtime(self):
        return math.isinf(self.game_timer_seconds)

    @property
    def is_kickoff(self):
        return self.kickoff_timer_seconds > 0


class ScoreboardProvider(SharedInfoProvider[AgentID, GameState]):
    """
    Implements a Rocket League scoreboard, since RLGym does not provide one.
    It tracks:
    - The current scoreline
    - The time remaining
    - The time remaining until kickoff buffer ends
    - Whether the game should go to kickoff
    - Whether the game is over
    Your state setter is responsible for initializing the scoreboard with the desired values.
    """

    def __init__(self):
        self.last_ticks = None
        self.info = None

    def create(self, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        if "scoreboard" not in shared_info:
            # Default behavior is infinite overtime, basically the same as no scoreboard
            shared_info["scoreboard"] = ScoreboardInfo(
                game_timer_seconds=math.inf,
                kickoff_timer_seconds=0.,
                blue_score=0,
                orange_score=0,
                go_to_kickoff=False,
                is_over=True,
            )
        return shared_info

    def set_state(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        info = shared_info.get("scoreboard")
        if info is None:
            raise ValueError("ScoreboardProvider requires a 'scoreboard' key in shared_info")

        if isinstance(info, dict):
            info = ScoreboardInfo(**info)

        assert not any(v is None for v in asdict(info).values()), "ScoreboardInfo must be fully defined"

        if info.is_over:
            raise ValueError("Cannot set scoreboard to be over")

        self.info = info

        self.last_ticks = initial_state.tick_count

        shared_info["scoreboard"] = info
        return shared_info

    def step(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        ticks_passed = state.tick_count - self.last_ticks
        self.last_ticks = state.tick_count

        # Copy info into new object to avoid modifying in place
        info: ScoreboardInfo = replace(self.info)

        if info.go_to_kickoff or info.is_over:
            ticks_passed = 0

        info.go_to_kickoff = False
        info.is_over = False

        if state.goal_scored:
            if state.scoring_team == BLUE_TEAM:
                info.blue_score += 1
            else:
                info.orange_score += 1
            if info.is_overtime:
                info.is_over = True
            info.go_to_kickoff = True

        if info.kickoff_timer_seconds > 0 and state.ball.position[1] == 0:
            info.kickoff_timer_seconds -= ticks_passed / TICKS_PER_SECOND
            if info.kickoff_timer_seconds <= 0:
                info.kickoff_timer_seconds = 0
        else:
            info.kickoff_timer_seconds = 0
            info.game_timer_seconds -= ticks_passed / TICKS_PER_SECOND

        if info.game_timer_seconds < 0:
            info.game_timer_seconds = 0
            if ball_hit_ground(ticks_passed, state.ball) or state.goal_scored:
                if info.blue_score == info.orange_score:
                    info.game_timer_seconds = math.inf
                    info.go_to_kickoff = True
                else:
                    info.is_over = True

        self.info = info
        shared_info["scoreboard"] = self.info
        return shared_info



from typing import Any, Dict, List

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import CAR_MAX_SPEED


class DemoReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, attacker_reward: float = 1.0, victim_punishment: float = 1.0,
                 bump_acceleration_reward: float = 0.0):
        self.attacker_reward = attacker_reward
        self.victim_punishment = victim_punishment
        self.bump_acceleration_reward = bump_acceleration_reward

        self.prev_state = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prev_state = initial_state

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = {agent: 0 for agent in agents}
        for agent in agents:
            car = state.cars[agent]
            victim = car.bump_victim_id
            if victim is not None:
                victim_car = state.cars[victim]
                if victim_car.is_demoed:
                    rewards[agent] += self.attacker_reward
                    rewards[victim] -= self.victim_punishment
                else:
                    acceleration = np.linalg.norm(state.cars[victim].physics.linear_velocity
                                                  - self.prev_state.cars[victim].physics.linear_velocity)
                    is_teammate = car.team_num == victim_car.team_num
                    reward = self.bump_acceleration_reward * acceleration / CAR_MAX_SPEED
                    rewards[agent] += reward if not is_teammate else -reward

        self.prev_state = state

        return rewards



from typing import Dict, Any

from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.state_mutators import KickoffMutator

from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


class GameMutator(StateMutator[GameState]):
    def __init__(self, game_length_seconds: float = 300.0, kickoff_timer_seconds: float = 5.0):
        self.game_length_seconds = game_length_seconds
        self.kickoff_timer_seconds = kickoff_timer_seconds
        self.kickoff_mutator = KickoffMutator()

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        scoreboard: ScoreboardInfo = shared_info["scoreboard"]
        if scoreboard.is_over:
            scoreboard.game_timer_seconds = self.game_length_seconds
            scoreboard.kickoff_timer_seconds = self.kickoff_timer_seconds
            scoreboard.blue_score = 0
            scoreboard.orange_score = 0
            scoreboard.go_to_kickoff = True
            scoreboard.is_over = False

        if scoreboard.go_to_kickoff:
            self.kickoff_mutator.apply(state, shared_info)

            assert state.ball.position[1] == 0, "Ball is not in kickoff position"

            scoreboard.kickoff_timer_seconds = self.kickoff_timer_seconds
            scoreboard.go_to_kickoff = False



from typing import List, Dict, Any

import numpy as np
from rlgym.api import RewardFunction, AgentID, StateType
from rlgym.rocket_league.api import GameState


class StackReward(RewardFunction[AgentID, GameState, np.ndarray]):
    """
    Turns the results from several reward functions into a single vector with each reward.
    Might be useful if someone wants to compute values and do some custom post-processing on them,
    e.g. adaptive weighting, normalization, logging.
    """
    def __init__(self, reward_functions: List[RewardFunction[AgentID, GameState, float]]):
        self.reward_functions = reward_functions

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        for reward_function in self.reward_functions:
            reward_function.reset(agents, initial_state, shared_info)

    def get_rewards(self, agents: List[AgentID], state: StateType, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, np.ndarray]:
        rewards = {agent: [] for agent in agents}
        for reward_function in self.reward_functions:
            for agent, reward in reward_function.get_rewards(agents, state, is_terminated, is_truncated,
                                                             shared_info).items():
                rewards[agent].append(reward)
        return {agent: np.stack(rewards[agent]) for agent in agents}



from typing import Generator

from rlgym.rocket_league.api import Car, PhysicsObject, GameState
from rlgym.rocket_league.common_values import BLUE_TEAM

from rlgym_tools.rocket_league.v1 import V1PhysicsObject, V1PlayerData, V1GameState


def convert_physics_object(po: PhysicsObject, copy: bool = True) -> V1PhysicsObject:
    pos = po.position
    quat = po.quaternion
    lin_vel = po.linear_velocity
    ang_vel = po.angular_velocity

    if copy:
        pos = pos.copy()
        quat = quat.copy()
        lin_vel = lin_vel.copy()
        ang_vel = ang_vel.copy()

    v1_po = V1PhysicsObject(
        position=pos,
        quaternion=quat,
        linear_velocity=lin_vel,
        angular_velocity=ang_vel,
    )

    return v1_po


def convert_car(car: Car, car_id: int = -1, copy: bool = True) -> V1PlayerData:
    pd = V1PlayerData()
    pd.car_id = car_id
    pd.team_num = car.team_num
    pd.match_goals = 0
    pd.match_saves = 0
    pd.match_shots = 0
    pd.match_demolishes = 0
    pd.boost_pickups = 0
    pd.is_demoed = car.is_demoed
    pd.on_ground = car.on_ground
    pd.ball_touched = car.ball_touches > 0
    pd.has_jump = not car.has_jumped
    pd.has_flip = car.has_flip
    pd.boost_amount = car.boost_amount
    pd.car_data = convert_physics_object(car.physics, copy=copy)
    pd.inverted_car_data = convert_physics_object(car.inverted_physics, copy=copy)

    return pd


def convert_game_state(state: GameState, copy: bool = True) -> V1GameState:
    v1_state = V1GameState()
    # v1_state.game_type = 0
    if state.goal_scored:
        if state.scoring_team == BLUE_TEAM:
            v1_state.blue_score += 1
        else:
            v1_state.orange_score += 1
    v1_state.last_touch = -1

    # Sort cars by (team, agent_id)
    sorted_cars = sorted(state.cars.items(), key=lambda x: x[0])
    blue_cars = []
    orange_cars = []
    for agent_id, car in sorted_cars:
        if car.team_num == BLUE_TEAM:
            blue_cars.append(car)
        else:
            orange_cars.append(car)
    blue_start_index = 1
    if len(blue_cars) <= 4 and len(orange_cars) <= 4:
        orange_start_index = 5
    else:
        orange_start_index = len(blue_cars) + 1

    v1_state.players = ([convert_car(car, car_id=blue_start_index + i, copy=copy)
                         for i, car in enumerate(blue_cars)]
                        + [convert_car(car, car_id=orange_start_index + i, copy=copy)
                           for i, car in enumerate(orange_cars)])

    v1_state.ball = convert_physics_object(state.ball, copy=copy)
    v1_state.inverted_ball = convert_physics_object(state.inverted_ball, copy=copy)

    v1_state.boost_pads[:] = state.boost_pad_timers == 0
    v1_state.inverted_boost_pads[:] = state.inverted_boost_pad_timers == 0

    return v1_state


def convert_game_states(copy: bool = True) -> Generator[GameState, V1GameState, None]:
    # Slightly more accurate as it tracks last touch, demos, goals and boost pickups.
    # Still no saves and shots though.
    # Usage:
    # game_state_converter = convert_game_states()
    # next(game_state_converter)  # Start the generator
    # Do these two lines for each subsequent state:
    #   game_state_converter.send(state)  # Send state
    #   v1_state = next(game_state_converter)
    latest_touch = None
    blue_goals = 0
    orange_goals = 0
    prev_boost_amounts = {}
    while True:
        state: GameState = yield
        v1_state = convert_game_state(state, copy=copy)
        cars = sorted(state.cars.items(), key=lambda x: x[0])
        for (agent_id, car), player_data in zip(cars, v1_state.players):
            # Demos
            if car.bump_victim_id is not None and state.cars[car.bump_victim_id].is_demoed:
                player_data.match_demolishes += 1

            # Last touch for goal tracking
            if car.ball_touches > 0:
                latest_touch = player_data.car_id

            # Boost pickups
            if car.is_demoed:
                prev_boost_amounts.pop(agent_id, None)
            else:
                if agent_id in prev_boost_amounts:
                    diff = car.boost_amount - prev_boost_amounts[agent_id]
                    if diff > 0:
                        player_data.boost_pickups += 1
                prev_boost_amounts[agent_id] = car.boost_amount

        # Goals
        if state.goal_scored:
            if state.scoring_team == BLUE_TEAM:
                blue_goals += 1
            else:
                orange_goals += 1
            for player_data in v1_state.players:
                if player_data.car_id == latest_touch:
                    player_data.match_goals += 1

        v1_state.blue_score = blue_goals
        v1_state.orange_score = orange_goals

        yield v1_state



import json
import os
import subprocess
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import pandas as pd

CARBALL_COMMAND = '{} -i "{}" -o "{}" parquet'

ENV = os.environ.copy()
ENV["NO_COLOR"] = "1"
ENV["RUST_BACKTRACE"] = "full"


def process_replay(replay_path, output_folder, carball_path=None, skip_existing=True):
    if carball_path is None:
        # Use carball.exe in the same directory as this script
        carball_path = os.path.join(os.path.dirname(__file__), "carball.exe")
    folder, fn = os.path.split(replay_path)
    replay_name = fn.replace(".replay", "")
    processed_folder = os.path.join(output_folder, replay_name)
    if os.path.isdir(processed_folder) and len(os.listdir(processed_folder)) > 0:
        if skip_existing:
            return
        else:
            os.rmdir(processed_folder)
    os.makedirs(processed_folder, exist_ok=True)

    with open(os.path.join(processed_folder, "carball.o.log"), "w", encoding="utf8") as stdout_f:
        with open(os.path.join(processed_folder, "carball.e.log"), "w", encoding="utf8") as stderr_f:
            return subprocess.run(
                CARBALL_COMMAND.format(carball_path, replay_path, processed_folder),
                stdout=stdout_f,
                stderr=stderr_f,
                env=ENV
            )


def load_parquet(*args, **kwargs):
    return pd.read_parquet(*args, engine="pyarrow", **kwargs)


@dataclass
class ParsedReplay:
    metadata: dict
    analyzer: dict
    game_df: pd.DataFrame
    ball_df: pd.DataFrame
    player_dfs: Dict[str, pd.DataFrame]

    @staticmethod
    def load(replay_dir, carball_path=None) -> "ParsedReplay":
        if not os.path.exists(replay_dir):
            raise FileNotFoundError(f"Replay directory {replay_dir} does not exist")

        if isinstance(replay_dir, str):
            replay_dir = Path(replay_dir)

        def load_files(rdir):
            try:
                with (rdir / "metadata.json").open("r", encoding="utf8") as f:
                    metadata = json.load(f)
                with (rdir / "analyzer.json").open("r", encoding="utf8") as f:
                    analyzer = json.load(f)
                ball_df = load_parquet(rdir / "__ball.parquet")
                game_df = load_parquet(rdir / "__game.parquet")

                player_dfs = {}
                for player_file in rdir.glob("player_*.parquet"):
                    player_id = player_file.name.split("_")[1].split(".")[0]
                    player_dfs[player_id] = load_parquet(player_file)

                return ParsedReplay(metadata, analyzer, game_df, ball_df, player_dfs)
            except FileNotFoundError as e:
                try:
                    with (rdir / "carball.e.log").open("r", encoding="utf8") as f:
                        raise ValueError(f"Error processing replay: {f.read()}")
                except FileNotFoundError:
                    raise e

        if not replay_dir.is_dir():
            # Assume it's a replay file
            with tempfile.TemporaryDirectory() as temp_dir:
                process_replay(replay_dir, temp_dir, carball_path=carball_path, skip_existing=False)
                replay_dir = Path(temp_dir) / replay_dir.stem
                return load_files(replay_dir)
        else:
            return load_files(replay_dir)



import math
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from rlgym.api import RewardFunction, SharedInfoProvider

from rlgym_tools.rocket_league.math.gamma import half_life_to_gamma
from rlgym_tools.rocket_league.replays.convert import replay_to_rlgym
from rlgym_tools.rocket_league.replays.parsed_replay import ParsedReplay


def get_reward_df(replay: ParsedReplay, reward_functions: List[Tuple[RewardFunction, str]],
                  shared_info_provider: Optional[SharedInfoProvider] = None):
    """
    Get a pandas dataframe of rewards for each player and reward function.

    :param replay: The replay to get rewards from.
    :param reward_functions: A list of tuples of reward functions and tags to separate them.
    :param shared_info_provider: A shared info provider to provide additional information to the reward functions.
    :return: A tuple of two dataframes: one for metadata (game time and frame number) and one for all the rewards.
    """
    reward_functions = {
        (reward_fn.__class__.__name__ + tag
         if not hasattr(reward_fn, "reward_fn")
         else reward_fn.reward_fn.__class__.__name__ + tag): reward_fn
        for reward_fn, tag in reward_functions
    }

    pid_to_name = {int(p["unique_id"]): p["name"]
                   for p in replay.metadata["players"]
                   if p["unique_id"] in replay.player_dfs}

    print(pid_to_name)

    # To be put into a pandas dataframe
    data = {
        "game_time": [],
        "frame_number": [],
        **{
            f"{player_name}/{reward_name}": []
            for reward_name in reward_functions.keys()
            for player_name in pid_to_name.values()
        }
    }

    mean_delta = replay.game_df["delta"].mean()
    shared_info = shared_info_provider.create({}) if shared_info_provider is not None else {}
    iterator = replay_to_rlgym(replay)
    reset = True
    for frame_number, replay_frame in enumerate(iterator):
        game_state = replay_frame.state
        agent_ids = sorted(game_state.cars.keys())

        if reset:
            if shared_info_provider is not None:
                shared_info_provider.set_state(agent_ids, game_state, shared_info)
            for reward_fn in reward_functions.values():
                reward_fn.reset(agent_ids, game_state, shared_info)
            reset = False

        if replay_frame.scoreboard.go_to_kickoff or replay_frame.scoreboard.is_over:
            reset = True

        terminated = {agent_id: reset for agent_id in agent_ids}
        truncated = {agent_id: False for agent_id in agent_ids}

        timer = replay_frame.scoreboard.game_timer_seconds
        if math.isinf(timer):
            prev_time = data["game_time"][-1] if data["game_time"] else 0
            data["game_time"].append(prev_time - mean_delta)
        else:
            data["game_time"].append(timer)
        data["frame_number"].append(frame_number)

        if shared_info_provider is not None:
            shared_info_provider.step(agent_ids, game_state, shared_info)
        shared_info["scoreboard"] = replay_frame.scoreboard
        for reward_name, reward_fn in reward_functions.items():
            reward = reward_fn.get_rewards(agent_ids, game_state, terminated, truncated, shared_info)
            for agent_id, r in reward.items():
                player_name = pid_to_name[agent_id]
                data[f"{player_name}/{reward_name}"].append(r)

    data = pd.DataFrame(data)

    meta = data[['game_time', 'frame_number']]
    multicol = data.loc[:, data.columns.str.contains("/")]
    # Make columns multiindex (player_name, reward_name)
    multicol.columns = pd.MultiIndex.from_tuples(
        [tuple(col.split("/", maxsplit=1)) for col in multicol.columns]
    )

    return meta, multicol


def generate_report(meta: pd.DataFrame, rewards: pd.DataFrame, per_player: bool = False):
    """
    Use the metadata and rewards dataframes to generate a report on the rewards.

    :param meta: The metadata dataframe.
    :param rewards: The rewards dataframe.
    :param per_player: Whether to generate a report for each player.
    :return:
    """
    if per_player:
        for player in rewards.columns.levels[0]:
            player_rewards = rewards[player]
            print(player)
            generate_report(meta, player_rewards)
            print()
    for reward_name in rewards.columns.levels[1]:
        reward_df = rewards.xs(reward_name, axis=1, level=1)

        # Flatten
        stacked = pd.concat([reward_df[col] for col in reward_df.columns], axis=0)
        num_players = len(stacked) // len(meta)
        times = np.repeat(meta['game_time'].values, num_players)

        print(reward_name)

        # First, some basic stats
        game_time = meta['game_time'].max() - meta['game_time'].min()
        t_tot = num_players * game_time
        print(f"\tPlayer count: {num_players}")
        print(f"\tGame length: {game_time:.2f} seconds")
        print("\tReward:")
        print(f"\t\tAverage (per-second): {stacked.sum() / t_tot:.3g}")
        print(f"\t\tMax: {stacked.max():.3g}")
        print(f"\t\tMin: {stacked.min():.3g}")

        print()

        # Next, some more intricate stats
        nonzero = stacked[stacked != 0]
        c = nonzero.count()
        density = c / stacked.count()
        unique = np.unique(nonzero.values)
        print("\tClassification:", end=" ")
        if density < 0.5:
            # sparsity = 1 - density
            regional_density = 1 - np.diff(stacked != 0).sum() / c
            density_str = f"{density:.1%}"
            if density_str == "0.0%":
                density_str = f"<0.1% ({c} in {stacked.count()})"
            if c < 10:
                print(f"sparse, with density={density_str}")
            elif regional_density > 0.5:
                print(f"regionally dense, with density={density_str} "
                      f"and regional density={regional_density:.3g})")
            else:
                print(f"sparse, with density={density_str}")
        else:
            print(f"dense, with density={density:.1%}")

        print()

        print("\tCount of non-zero rewards:", c)
        print("\tAverage of non-zero rewards:", nonzero.mean())
        print("\tCount of positive rewards:", (nonzero > 0).sum())
        print("\tAverage of positive rewards:", nonzero[nonzero > 0].mean())
        print("\tCount of negative rewards:", (nonzero < 0).sum())
        print("\tAverage of negative rewards:", nonzero[nonzero < 0].mean())
        if len(unique) <= 5 and c > len(unique):
            print("\tUnique values:")
            for val in unique:
                if np.isnan(val):
                    continue
                count = (nonzero == val).sum()
                print(f"\t\t{val}: {count} occurrences")

        print()


def generate_plot(meta_df, reward_df, gamma=None):
    """
    Generate a plot of the rewards for each player.

    :param meta_df: The metadata dataframe.
    :param reward_df: The rewards dataframe.
    :param gamma: The discount factor for the discounted future sum.
    :return:
    """
    if gamma is None:
        gamma = half_life_to_gamma(30, 4)
    xs = meta_df["game_time"]
    # Make a plot of each reward for each player
    fig, axs = plt.subplots(nrows=len(reward_df.columns.levels[1]),
                            ncols=len(reward_df.columns.levels[0]),
                            figsize=(10 * len(reward_df.columns.levels[0]), 5 * len(reward_df.columns.levels[1])),
                            sharey="row")
    goals = np.where(reward_df[reward_df.columns.levels[0][0], "EpisodeEnd"] != 0)[0]
    edges = sorted(set([0] + (goals + 1).tolist() + [len(xs)]))
    for i, player in enumerate(reward_df.columns.levels[0]):
        for j, reward in enumerate(reward_df.columns.levels[1]):
            ax = axs[j, i]
            # Cumulative reward
            tot = pd.concat([reward_df[player, reward].iloc[start:end].cumsum()
                             for start, end in zip(edges, edges[1:])]).values
            ax.plot(xs, tot, label="cumulative")
            # Gamma discounted cumulative reward
            disc_sum = pd.concat([reward_df[player, reward].iloc[start:end][::-1].ewm(alpha=1 - gamma).sum()[::-1]
                                  for start, end in zip(edges, edges[1:])]).values
            ax.plot(xs, disc_sum, linestyle="--", label="discounted future sum")
            # Vertical lines for goals
            for g, goal in enumerate(goals):
                ax.axvline(xs[goal], color="red", linestyle="--", alpha=0.5, label="goal" if g == 0 else None)
            ax.legend()
            ax.set_ylabel("Cumulative reward")
            ax.set_title(f"{player}/{reward}")
            ax.grid(True)

            # Invert x-axis for time
            ax.invert_xaxis()
            ax.set_xlabel("Time (s)")

            # Make secondary x-axis tick labels that show mm:ss
            a_sec = ax.secondary_xaxis("top")
            a_sec.set_xlabel("Time (mm:ss)")
            a_sec.set_xticks(ax.get_xticks())
            a_sec.set_xticklabels([f"{int(t // 60):02d}:{int(t % 60):02d}" for t in ax.get_xticks()])
    fig.tight_layout()
    return fig



import math

from rlgym.rocket_league.common_values import TICKS_PER_SECOND


# Utility functions for converting between gamma, half-life, and horizon.

def half_life_to_gamma(half_life_seconds: float, tick_skip: int) -> float:
    half_life_ticks = half_life_seconds * TICKS_PER_SECOND
    half_life_steps = half_life_ticks / tick_skip
    gamma = 0.5 ** (1 / half_life_steps)
    # gamma^half_life_steps = 0.5
    return gamma


def gamma_to_half_life(gamma: float, tick_skip: int) -> float:
    half_life_steps = -math.log(2) / math.log(gamma)
    half_life_ticks = half_life_steps * tick_skip
    half_life_seconds = half_life_ticks / TICKS_PER_SECOND
    return half_life_seconds


def horizon_to_gamma(horizon_seconds: float, tick_skip: int) -> float:
    horizon_ticks = horizon_seconds * TICKS_PER_SECOND
    horizon_steps = horizon_ticks / tick_skip
    gamma = 1 - 1 / horizon_steps
    # 1/(1-gamma) = horizon_steps
    return gamma


def gamma_to_horizon(gamma: float, tick_skip: int) -> float:
    horizon_steps = 1 / (1 - gamma)
    horizon_ticks = horizon_steps * tick_skip
    horizon_seconds = horizon_ticks / TICKS_PER_SECOND
    return horizon_seconds



from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


@dataclass(slots=True)
class ReplayFrame:
    state: GameState  # Current state of the game
    actions: Dict[int, np.ndarray]  # Actions for each player
    update_age: Dict[int, float]  # Time since the replay updated values for each car (>0 means state is interpolated)
    scoreboard: ScoreboardInfo  # Current scoreboard
    episode_seconds_remaining: float  # Time remaining until someone scores or ball hits ground at 0s
    next_scoring_team: Optional[int]  # Team that scores the next goal, None if ball hits ground at 0s
    winning_team: Optional[int]  # Team that wins the game, None if game ended without a winner



from typing import List

import numpy as np
from rlgym.rocket_league.api import PhysicsObject


def relative_physics(origin: PhysicsObject, targets: List[PhysicsObject]) -> List[PhysicsObject]:
    result = []
    rot = origin.rotation_mtx
    for target in targets:
        po = PhysicsObject()
        po.position = (target.position - origin.position) @ rot
        po.linear_velocity = (target.linear_velocity - origin.linear_velocity) @ rot
        po.angular_velocity = target.angular_velocity @ rot
        po.rotation_mtx = target.rotation_mtx @ rot.T
        result.append(po)
    return result


def dodge_relative_rot_mtx(rot_mtx: np.ndarray) -> np.ndarray:
    modified_rot = np.zeros_like(rot_mtx)
    fw = rot_mtx[:2, 0]
    modified_rot[:2, 0] = fw / np.linalg.norm(fw)  # Renormalize forward
    modified_rot[0, 1] = -modified_rot[1, 0]  # Recalculate right vector
    modified_rot[1, 1] = modified_rot[0, 0]  # --||--
    modified_rot[2, 2] = 1  # Set z axis to be up
    return modified_rot


def dodge_relative_physics(origin: PhysicsObject, targets: List[PhysicsObject]) -> List[PhysicsObject]:
    # Dodges only happen in the xy plane, so we pretend the car's hood is facing straight up
    modified_rot = dodge_relative_rot_mtx(origin.rotation_mtx)
    result = []
    for target in targets:
        po = PhysicsObject()
        po.position = (target.position - origin.position) @ modified_rot
        po.linear_velocity = (target.linear_velocity - origin.linear_velocity) @ modified_rot
        po.angular_velocity = target.angular_velocity @ modified_rot
        po.rotation_mtx = target.rotation_mtx @ modified_rot.T
        result.append(po)

    return result



from typing import Literal

import numpy as np
from rlgym.rocket_league.api import GameState, Car, PhysicsObject, GameConfig
from rlgym.rocket_league.common_values import BOOST_LOCATIONS

from rlgym_tools.rocket_league.replays.replay_frame import ReplayFrame
from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


# Utilities to serialize and deserialize RLGym Rocket League objects into numpy arrays
# This can be useful for saving and loading states, sending over network, or batch operations.


def serialize(obj):
    if isinstance(obj, GameState):
        return serialize_game_state(obj)
    elif isinstance(obj, GameConfig):
        return serialize_config(obj)
    elif isinstance(obj, Car):
        return serialize_car(obj)
    elif isinstance(obj, PhysicsObject):
        return serialize_physics_object(obj)
    elif isinstance(obj, ScoreboardInfo):
        return serialize_scoreboard(obj)
    elif isinstance(obj, ReplayFrame):
        return serialize_replay_frame(obj)
    else:
        raise ValueError(f"Unsupported type: {type(obj)}")


def deserialize(obj):
    if isinstance(obj, np.ndarray):
        if obj.shape == (18,):
            return deserialize_physics_object(obj)
        elif obj.shape == (45,):
            return deserialize_car(obj)
        elif obj.shape == (3,):
            return deserialize_config(obj)
        elif obj.shape == (6,):
            return deserialize_scoreboard(obj)
        else:
            try:
                return deserialize_game_state(obj)
            except AssertionError:
                return deserialize_replay_frame(obj)
    else:
        raise ValueError(f"Unsupported type: {type(obj)}")


# Indices for serialized GameState:
GS_TICK_COUNT = 0
GS_GOAL_SCORED = 1
GS_CONFIG = slice(2, 5)
GS_BALL = slice(5, 23)
GS_BOOST_PAD_TIMERS = slice(23, 23 + len(BOOST_LOCATIONS))
GS_CARS = slice(23 + len(BOOST_LOCATIONS), None)
GS_CAR_LENGTH = 46  # 1 for agent ID, 45 for car


def serialize_game_state(game_state: GameState, aid_to_num=None) -> np.ndarray:
    # Make sure agent IDs are compatible with serialization
    if aid_to_num is None:
        aid_to_num = {}
        for agent_id in sorted(game_state.cars.keys()):
            if isinstance(agent_id, (float, int)):
                aid_to_num[agent_id] = agent_id
            else:
                aid_to_num[agent_id] = len(aid_to_num)

    s = np.zeros(2 + 3 + 18 + len(BOOST_LOCATIONS) + len(game_state.cars) * GS_CAR_LENGTH,
                 dtype=np.float32)
    s[GS_TICK_COUNT] = game_state.tick_count
    s[GS_GOAL_SCORED] = game_state.goal_scored
    s[GS_CONFIG] = serialize_config(game_state.config)
    s[GS_BALL] = serialize_physics_object(game_state.ball)
    s[GS_BOOST_PAD_TIMERS] = game_state.boost_pad_timers
    for i, (aid, car) in enumerate(game_state.cars.items()):
        start = GS_CARS.start + i * GS_CAR_LENGTH
        end = start + GS_CAR_LENGTH
        s[start] = aid_to_num[aid]
        s[start + 1:end] = serialize_car(car, aid_to_num)
    return s


def deserialize_game_state(data: np.ndarray) -> GameState:
    assert (len(data) - GS_CARS.start) % GS_CAR_LENGTH == 0, f"Invalid data length: {len(data)}"
    game_state = GameState()
    game_state.tick_count = int(data[GS_TICK_COUNT])
    game_state.goal_scored = bool(data[GS_GOAL_SCORED])

    config = deserialize_config(data[GS_CONFIG])
    game_state.config = config

    ball = deserialize_physics_object(data[GS_BALL])
    game_state.ball = ball

    game_state.boost_pad_timers = data[GS_BOOST_PAD_TIMERS]

    rest = data[GS_CARS]
    cars = {}
    for i in range(0, len(rest), GS_CAR_LENGTH):
        if np.all(rest[i:i + GS_CAR_LENGTH] == 0):  # Padding
            break
        agent_id = int(rest[i])
        car = deserialize_car(rest[i + 1:i + GS_CAR_LENGTH])
        cars[agent_id] = car
    game_state.cars = cars

    return game_state


# Indices for serialized GameConfig:
CONFIG_GRAVITY = 0
CONFIG_BOOST_CONSUMPTION = 1
CONFIG_DODGE_DEADZONE = 2


def serialize_config(config: GameConfig) -> np.ndarray[tuple[Literal[3]], np.dtype[np.float32]]:
    s = np.zeros(3, dtype=np.float32)
    s[CONFIG_GRAVITY] = config.gravity
    s[CONFIG_BOOST_CONSUMPTION] = config.boost_consumption
    s[CONFIG_DODGE_DEADZONE] = config.dodge_deadzone
    return s


def deserialize_config(data: np.ndarray[tuple[Literal[3]], np.dtype[np.float32]]) -> GameConfig:
    assert data.shape == (3,)
    config = GameConfig()
    config.gravity = float(data[CONFIG_GRAVITY])
    config.boost_consumption = float(data[CONFIG_BOOST_CONSUMPTION])
    config.dodge_deadzone = float(data[CONFIG_DODGE_DEADZONE])

    return config


# Indices for serialized Car:
CAR_TEAM_NUM = 0
CAR_HITBOX_TYPE = 1
CAR_BALL_TOUCHES = 2
CAR_BUMP_VICTIM_ID = 3

CAR_DEMO_RESPAWN_TIMER = 4
CAR_WHEELS_WITH_CONTACT = range(5, 9)
CAR_SUPERSONIC_TIME = 9
CAR_BOOST_AMOUNT = 10
CAR_BOOST_ACTIVE_TIME = 11
CAR_HANDBRAKE = 12

CAR_HAS_JUMPED = 13
CAR_IS_HOLDING_JUMP = 14
CAR_IS_JUMPING = 15
CAR_JUMP_TIME = 16

CAR_HAS_FLIPPED = 17
CAR_HAS_DOUBLE_JUMPED = 18
CAR_AIR_TIME_SINCE_JUMP = 19
CAR_FLIP_TIME = 20
CAR_FLIP_TORQUE = slice(21, 24)

CAR_IS_AUTOFLIPPING = 24
CAR_AUTOFLIP_TIMER = 25
CAR_AUTOFLIP_DIRECTION = 26

CAR_PHYSICS = slice(27, 27 + 18)


def serialize_car(car: Car, bump_victim_map=None) -> np.ndarray[tuple[Literal[42]], np.dtype[np.float32]]:
    bump_victim = car.bump_victim_id
    if bump_victim is None:
        bump_victim = -1
    else:
        bump_victim_map = bump_victim_map or {}
        if not isinstance(bump_victim, (int, float)):
            bump_victim = bump_victim_map.get(bump_victim, -2)
        assert bump_victim != -1, "Invalid bump victim ID (-1 is reserved for None)"

    s = np.zeros(45, dtype=np.float32)
    s[CAR_TEAM_NUM] = car.team_num
    s[CAR_HITBOX_TYPE] = car.hitbox_type
    s[CAR_BALL_TOUCHES] = car.ball_touches
    s[CAR_BUMP_VICTIM_ID] = bump_victim
    s[CAR_DEMO_RESPAWN_TIMER] = car.demo_respawn_timer
    s[CAR_WHEELS_WITH_CONTACT] = car.wheels_with_contact
    s[CAR_SUPERSONIC_TIME] = car.supersonic_time
    s[CAR_BOOST_AMOUNT] = car.boost_amount
    s[CAR_BOOST_ACTIVE_TIME] = car.boost_active_time
    s[CAR_HANDBRAKE] = car.handbrake
    s[CAR_HAS_JUMPED] = car.has_jumped
    s[CAR_IS_HOLDING_JUMP] = car.is_holding_jump
    s[CAR_IS_JUMPING] = car.is_jumping
    s[CAR_JUMP_TIME] = car.jump_time
    s[CAR_HAS_FLIPPED] = car.has_flipped
    s[CAR_HAS_DOUBLE_JUMPED] = car.has_double_jumped
    s[CAR_AIR_TIME_SINCE_JUMP] = car.air_time_since_jump
    s[CAR_FLIP_TIME] = car.flip_time
    s[CAR_FLIP_TORQUE] = car.flip_torque
    s[CAR_IS_AUTOFLIPPING] = car.is_autoflipping
    s[CAR_AUTOFLIP_TIMER] = car.autoflip_timer
    s[CAR_AUTOFLIP_DIRECTION] = car.autoflip_direction
    s[CAR_PHYSICS] = serialize_physics_object(car.physics)
    return s


def deserialize_car(data: np.ndarray[tuple[Literal[42]], np.dtype[np.float32]]) -> Car:
    assert data.shape == (45,)
    car = Car()
    car.team_num = int(data[CAR_TEAM_NUM])
    car.hitbox_type = int(data[CAR_HITBOX_TYPE])
    car.ball_touches = int(data[CAR_BALL_TOUCHES])
    car.bump_victim_id = int(data[CAR_BUMP_VICTIM_ID]) if data[CAR_BUMP_VICTIM_ID] != -1 else None

    car.demo_respawn_timer = float(data[CAR_DEMO_RESPAWN_TIMER])
    car.wheels_with_contact = tuple(bool(data[i]) for i in CAR_WHEELS_WITH_CONTACT)  # noqa trust that length is correct
    car.supersonic_time = float(data[CAR_SUPERSONIC_TIME])
    car.boost_amount = float(data[CAR_BOOST_AMOUNT])
    car.boost_active_time = float(data[CAR_BOOST_ACTIVE_TIME])
    car.handbrake = float(data[CAR_HANDBRAKE])

    car.has_jumped = bool(data[CAR_HAS_JUMPED])
    car.is_holding_jump = bool(data[CAR_IS_HOLDING_JUMP])
    car.is_jumping = bool(data[CAR_IS_JUMPING])
    car.jump_time = float(data[CAR_JUMP_TIME])

    car.has_flipped = bool(data[CAR_HAS_FLIPPED])
    car.has_double_jumped = bool(data[CAR_HAS_DOUBLE_JUMPED])
    car.air_time_since_jump = float(data[CAR_AIR_TIME_SINCE_JUMP])
    car.flip_time = float(data[CAR_FLIP_TIME])
    car.flip_torque = data[CAR_FLIP_TORQUE]

    car.is_autoflipping = bool(data[CAR_IS_AUTOFLIPPING])
    car.autoflip_timer = float(data[CAR_AUTOFLIP_TIMER])
    car.autoflip_direction = float(data[CAR_AUTOFLIP_DIRECTION])

    car.physics = deserialize_physics_object(data[CAR_PHYSICS])

    return car


# Indices for serialized PhysicsObject:
PO_POSITION = slice(0, 3)
PO_LINEAR_VELOCITY = slice(3, 6)
PO_ANGULAR_VELOCITY = slice(6, 9)
PO_ROTATION_MTX = slice(9, 18)


def serialize_physics_object(physics_object: PhysicsObject) -> np.ndarray[tuple[Literal[18]], np.dtype[np.float32]]:
    s = np.zeros(18, dtype=np.float32)
    s[PO_POSITION] = physics_object.position
    s[PO_LINEAR_VELOCITY] = physics_object.linear_velocity
    s[PO_ANGULAR_VELOCITY] = physics_object.angular_velocity
    s[PO_ROTATION_MTX] = physics_object.rotation_mtx.flatten()
    return s


def deserialize_physics_object(data: np.ndarray[tuple[Literal[18]], np.dtype[np.float32]]) -> PhysicsObject:
    assert data.shape == (18,)
    po = PhysicsObject()
    po.position = data[PO_POSITION]
    po.linear_velocity = data[PO_LINEAR_VELOCITY]
    po.angular_velocity = data[PO_ANGULAR_VELOCITY]
    po.rotation_mtx = data[PO_ROTATION_MTX].reshape(3, 3)

    return po


# Indices for serialized ScoreboardInfo:
SB_GAME_TIMER_SECONDS = 0
SB_KICKOFF_TIMER_SECONDS = 1
SB_BLUE_SCORE = 2
SB_ORANGE_SCORE = 3
SB_GO_TO_KICKOFF = 4
SB_IS_OVER = 5


def serialize_scoreboard(scoreboard: ScoreboardInfo):
    s = np.zeros(6, dtype=np.float32)
    s[SB_GAME_TIMER_SECONDS] = scoreboard.game_timer_seconds
    s[SB_KICKOFF_TIMER_SECONDS] = scoreboard.kickoff_timer_seconds
    s[SB_BLUE_SCORE] = scoreboard.blue_score
    s[SB_ORANGE_SCORE] = scoreboard.orange_score
    s[SB_GO_TO_KICKOFF] = scoreboard.go_to_kickoff
    s[SB_IS_OVER] = scoreboard.is_over
    return s


def deserialize_scoreboard(data: np.ndarray):
    assert data.shape == (6,)
    return ScoreboardInfo(
        game_timer_seconds=float(data[SB_GAME_TIMER_SECONDS]),
        kickoff_timer_seconds=float(data[SB_KICKOFF_TIMER_SECONDS]),
        blue_score=int(data[SB_BLUE_SCORE]),
        orange_score=int(data[SB_ORANGE_SCORE]),
        go_to_kickoff=bool(data[SB_GO_TO_KICKOFF]),
        is_over=bool(data[SB_IS_OVER]),
    )


# Indices for serialized ReplayFrame:
RF_SCOREBOARD = slice(0, 6)
RF_EPISODE_SECONDS_REMAINING = 6
RF_NEXT_SCORING_TEAM = 7
RF_WINNING_TEAM = 8
RF_NUM_PLAYERS = 9
RF_AGENT_IDS_START = 10
RF_ACTION_SIZE = 8


def serialize_replay_frame(replay_frame: ReplayFrame):
    # state: GameState
    # actions: Dict[int, np.ndarray]
    # update_age: Dict[int, float]
    # scoreboard: ScoreboardInfo
    # episode_seconds_remaining: float
    # next_scoring_team: Optional[int]
    # winning_team: Optional[int]
    aid_to_num = {}
    agent_ids = sorted(replay_frame.state.cars.keys())
    for agent_id in agent_ids:
        if isinstance(agent_id, (float, int)):
            aid_to_num[agent_id] = agent_id
        else:
            aid_to_num[agent_id] = len(aid_to_num)

    num_players = len(replay_frame.state.cars)
    update_ages = np.array([replay_frame.update_age[agent_id] for agent_id in agent_ids], dtype=np.float32)
    actions = np.array([replay_frame.actions[agent_id] for agent_id in agent_ids], dtype=np.float32)
    agent_ids = np.array([aid_to_num[agent_id] for agent_id in agent_ids], dtype=np.float32)

    serialized_game_state = serialize_game_state(replay_frame.state, aid_to_num)
    s = np.zeros(6 + 3 + 1 + num_players + num_players + num_players * RF_ACTION_SIZE + serialized_game_state.shape[0],
                 dtype=np.float32)
    s[RF_SCOREBOARD] = serialize_scoreboard(replay_frame.scoreboard)
    s[RF_EPISODE_SECONDS_REMAINING] = replay_frame.episode_seconds_remaining
    s[RF_NEXT_SCORING_TEAM] = replay_frame.next_scoring_team if replay_frame.next_scoring_team is not None else -1
    s[RF_WINNING_TEAM] = replay_frame.winning_team if replay_frame.winning_team is not None else -1
    s[RF_NUM_PLAYERS] = num_players

    i = RF_AGENT_IDS_START
    j = i + num_players
    s[i:j] = agent_ids
    i = j
    j = i + num_players
    s[i:j] = update_ages
    i = j
    j = i + num_players * RF_ACTION_SIZE
    s[i:j] = actions.flatten()
    i = j
    j = i + serialized_game_state.shape[0]
    s[i:j] = serialized_game_state
    return s


def deserialize_replay_frame(data: np.ndarray):
    scoreboard = deserialize_scoreboard(data[RF_SCOREBOARD])
    episode_seconds_remaining = float(data[RF_EPISODE_SECONDS_REMAINING])
    next_scoring_team = int(data[RF_NEXT_SCORING_TEAM]) if data[RF_NEXT_SCORING_TEAM] != -1 else None
    winning_team = int(data[RF_WINNING_TEAM]) if data[RF_WINNING_TEAM] != -1 else None
    num_players = int(data[RF_NUM_PLAYERS])
    k = RF_AGENT_IDS_START
    agent_ids = data[k:k + num_players]
    agent_ids = [int(agent_id) for agent_id in agent_ids]
    k += num_players
    update_ages = data[k:k + num_players]
    k += num_players
    actions = data[k:k + num_players * RF_ACTION_SIZE].reshape(num_players, RF_ACTION_SIZE)
    k += num_players * RF_ACTION_SIZE
    state = deserialize_game_state(data[k:k + GS_CARS.start + num_players * GS_CAR_LENGTH])
    # assert np.all(data[k + GS_CARS.start + num_players * GS_CAR_LENGTH:] == 0), "Invalid padding"

    return ReplayFrame(
        state=state,
        actions={agent_id: action for agent_id, action in zip(agent_ids, actions)},
        update_age={agent_id: update_age for agent_id, update_age in zip(agent_ids, update_ages)},
        scoreboard=scoreboard,
        episode_seconds_remaining=episode_seconds_remaining,
        next_scoring_team=next_scoring_team,
        winning_team=winning_team,
    )



from typing import Any, Dict, List

from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState


class TeamSpiritRewardWrapper(RewardFunction[AgentID, GameState, float]):
    """
    Implements team spirit, as described in the OpenAI Five paper (Berner et al., 2019).
    """
    def __init__(self, reward_fn: RewardFunction, team_spirit: float):
        self.reward_fn = reward_fn
        self.team_spirit = team_spirit

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.reward_fn.reset(agents, initial_state, shared_info)

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        base_rewards = self.reward_fn.get_rewards(list(state.cars.keys()), state, is_terminated, is_truncated, shared_info)

        # We calculate the means first
        total_blue = total_orange = n_blue = n_orange = 0
        for agent, car in state.cars.items():
            if car.is_blue:
                total_blue += base_rewards[agent]
                n_blue += 1
            else:
                total_orange += base_rewards[agent]
                n_orange += 1
        mean_blue = total_blue / n_blue if n_blue > 0 else 0
        mean_orange = total_orange / n_orange if n_orange > 0 else 0

        # Then do the team spirit adjustment
        spirit_rewards = {}
        for agent in agents:
            rew = base_rewards[agent]
            if state.cars[agent].is_blue:
                team_mean, opp_mean = mean_blue, mean_orange
            else:
                team_mean, opp_mean = mean_orange, mean_blue
            # Actual formula here
            spirit_rewards[agent] = (1 - self.team_spirit) * rew + self.team_spirit * team_mean - opp_mean

        return spirit_rewards



from typing import List, Dict, Any

from rlgym.api import DoneCondition, AgentID
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.math.ball import ball_hit_ground


class BallHitGroundCondition(DoneCondition[AgentID, GameState]):
    def __init__(self):
        self.last_tick_count = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.last_tick_count = initial_state.tick_count

    def is_done(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[AgentID, bool]:
        ticks_passed = state.tick_count - self.last_tick_count
        done = ball_hit_ground(ticks_passed, state.ball, pre=False)

        self.last_tick_count = state.tick_count

        return {agent: done for agent in agents}



import math
from typing import List, Dict, Any, Callable

from rlgym.api import RewardFunction, AgentID, StateType, RewardType
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import TICKS_PER_SECOND


class BoostKeepReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, reward_per_second: float = 1.0,
                 activation_fn: Callable[[float], float] = lambda x: math.sqrt(0.01 * x)):
        """
        Reward function that rewards agents for having boost in their tank.

        :param reward_per_second: Amount of reward to give per second at full boost.
        :param activation_fn: Activation function to apply to the boost value before calculating the reward. Default is
                              the square root function so that increasing boost is more important when boost is low.
        """
        self.reward_per_tick = reward_per_second / TICKS_PER_SECOND
        self.activation_fn = activation_fn

        self.prev_ticks = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prev_ticks = initial_state.tick_count

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, RewardType]:
        ticks_passed = state.tick_count - self.prev_ticks
        mul = self.reward_per_tick * ticks_passed
        rewards = {}
        for agent in agents:
            boost = state.cars[agent].boost_amount
            rewards[agent] = self.activation_fn(boost) * mul
        self.prev_ticks = state.tick_count

        return rewards



import math
from typing import List, Dict, Any, Tuple

import numpy as np

from rlgym.api import ObsBuilder, AgentID
from rlgym.rocket_league.api import Car, GameState, PhysicsObject
from rlgym.rocket_league.common_values import ORANGE_TEAM
from rlgym_tools.rocket_league.math.relative import relative_physics, dodge_relative_physics


class RelativeDefaultObs(ObsBuilder[AgentID, np.ndarray, GameState, Tuple[str, int]]):
    """
    Relative version of the default observation builder.
    Agent's car uses absolute reference frame, other cars and ball are relative to the agent's car.
    """

    def __init__(self, zero_padding=3, pos_coef=1 / 2300, ang_coef=1 / math.pi, lin_vel_coef=1 / 2300,
                 ang_vel_coef=1 / math.pi,
                 pad_timer_coef=1 / 10, boost_coef=1 / 100, dodge_relative=False):
        """
        :param zero_padding: Number of max cars per team, if not None the obs will be zero padded
        :param pos_coef: Position normalization coefficient
        :param ang_coef: Rotation angle normalization coefficient
        :param lin_vel_coef: Linear velocity normalization coefficient
        :param ang_vel_coef: Angular velocity normalization coefficient
        :param pad_timer_coef: Boost pad timers normalization coefficient
        """
        super().__init__()
        self.POS_COEF = pos_coef
        self.ANG_COEF = ang_coef
        self.LIN_VEL_COEF = lin_vel_coef
        self.ANG_VEL_COEF = ang_vel_coef
        self.PAD_TIMER_COEF = pad_timer_coef
        self.BOOST_COEF = boost_coef
        self.zero_padding = zero_padding
        self.relative_fn = relative_physics if not dodge_relative else dodge_relative_physics

    def get_obs_space(self, agent: AgentID) -> Tuple[str, int]:
        if self.zero_padding is not None:
            return 'real', 52 + 20 * self.zero_padding * 2
        else:
            return 'real', -1  # Without zero padding this depends on the initial state, but we don't want to crash for now

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        pass

    def build_obs(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[
        AgentID, np.ndarray]:
        obs = {}
        for agent in agents:
            obs[agent] = self._build_obs(agent, state, shared_info)

        return obs

    def _build_obs(self, agent: AgentID, state: GameState, shared_info: Dict[str, Any]) -> np.ndarray:
        car = state.cars[agent]
        if car.team_num == ORANGE_TEAM:
            physics = car.inverted_physics
            inverted = True
            ball = state.inverted_ball
            pads = state.inverted_boost_pad_timers
        else:
            physics = car.physics
            inverted = False
            ball = state.ball
            pads = state.boost_pad_timers

        rel_ball, *rel_cars = relative_physics(car.physics,
                                               [state.ball] + [car.physics for car in state.cars.values()])

        obs = [  # Global stuff
            rel_ball.position * self.POS_COEF,
            rel_ball.linear_velocity * self.LIN_VEL_COEF,
            rel_ball.angular_velocity * self.ANG_VEL_COEF,
            pads * self.PAD_TIMER_COEF,
            [  # Partially observable variables
                car.is_holding_jump,
                car.handbrake,
                car.has_jumped,
                car.is_jumping,
                car.has_flipped,
                car.is_flipping,
                car.has_double_jumped,
                car.can_flip,
                car.air_time_since_jump
            ]
        ]

        car_obs = self._make_player_obs(physics, car)
        obs.append(car_obs)

        allies = []
        enemies = []

        for (other, other_car), other_physics in zip(state.cars.items(), rel_cars):
            if other == agent:
                continue

            if other_car.team_num == car.team_num:
                team_obs = allies
            else:
                team_obs = enemies

            team_obs.append(self._make_player_obs(other_physics, other_car))

        if self.zero_padding is not None:
            # Padding for multi game mode
            while len(allies) < self.zero_padding - 1:
                allies.append(np.zeros_like(car_obs))
            while len(enemies) < self.zero_padding:
                enemies.append(np.zeros_like(car_obs))

        obs.extend(allies)
        obs.extend(enemies)
        return np.concatenate(obs)

    def _make_player_obs(self, physics: PhysicsObject, car: Car) -> np.ndarray:
        return np.concatenate([
            physics.position * self.POS_COEF,
            physics.forward,
            physics.up,
            physics.linear_velocity * self.LIN_VEL_COEF,
            physics.angular_velocity * self.ANG_VEL_COEF,
            [car.boost_amount * self.BOOST_COEF,
             car.demo_respawn_timer,
             int(car.on_ground),
             int(car.is_boosting),
             int(car.is_supersonic)]
        ])



from typing import Any

import numpy as np
from rlgym.rocket_league.action_parsers import LookupTableAction


def _parse_bin(b, endpoint=True):
    if isinstance(b, int):
        b = np.linspace(-1, 1, b, endpoint=endpoint)
    else:
        b = np.array(b)
    return b


def _subdivide(lo, hi, depth=0):
    # Add points to a grid of size [lo, hi] x [lo, hi]
    # alternating between square and diamond steps as in the diamond-square algorithm
    # Basically, if we have:
    # •   •   •
    #
    # •   •   •
    #
    # •   •   •
    # then we add points to the grid like this:
    # •   •   •
    #   •   •
    # •   •   •
    #   •   •
    # •   •   •
    # and then like this:
    # • • • • •
    # • • • • •
    # • • • • •
    # • • • • •
    # • • • • •
    # instead of going straight from the first grid to the third grid
    if depth < 0:
        return
    # Square step
    count = 1 + 2 ** (depth // 2)
    bins, delta = np.linspace(lo, hi, count, retstep=True)
    for nx in bins:
        for ny in bins:
            yield nx, ny
    # Diamond step, just shift the square step diagonally and ignore those that exceed hi
    if depth % 2 == 1:
        bins = bins[:-1] + delta / 2
        for nx in bins:
            for ny in bins:
                yield nx, ny


class AdvancedLookupTableAction(LookupTableAction):
    def __init__(self, throttle_bins: Any = 3,
                 steer_bins: Any = 3,
                 torque_subdivisions: Any = 2,
                 flip_bins: Any = 8,
                 include_stalls: bool = False):
        super().__init__()
        self._lookup_table = self.make_lookup_table(throttle_bins, steer_bins, torque_subdivisions, flip_bins,
                                                    include_stalls)

    @staticmethod
    def make_lookup_table(throttle_bins: Any = 3,
                          steer_bins: Any = 3,
                          torque_subdivisions: Any = 2,
                          flip_bins: Any = 8,
                          include_stalls: bool = False):
        # Parse bins
        throttle_bins = _parse_bin(throttle_bins)
        steer_bins = _parse_bin(steer_bins)
        flip_bins = (_parse_bin(flip_bins,
                                endpoint=False) + 1) * np.pi  # Split a circle into equal segments in [0, 2pi)
        if isinstance(torque_subdivisions, int):
            torque_face = np.array([
                [x, y]
                for x, y in _subdivide(-1, 1, torque_subdivisions)
            ])
        else:
            if isinstance(torque_subdivisions, np.ndarray) and torque_subdivisions.ndim == 2:
                torque_face = torque_subdivisions
            else:
                torque_subdivisions = _parse_bin(torque_subdivisions)
                torque_face = np.array([
                    [x, y]
                    for x in torque_subdivisions
                    for y in torque_subdivisions
                ])

        actions = []

        # Ground
        pitch = roll = jump = 0
        for throttle in throttle_bins:
            for steer in steer_bins:
                for boost in (0, 1):
                    for handbrake in (0, 1):
                        if boost == 1 and throttle != 1:
                            continue
                        yaw = steer
                        actions.append([throttle, steer, pitch, yaw, roll, jump, boost, handbrake])

        # Aerial
        jump = handbrake = 0
        points = np.array([
            np.insert(p, i, side)
            for i in range(3)  # Determines which axis we select faces from
            for side in (-1, 1)  # Determines which side we select
            for p in torque_face  # Selects where we are on the face
        ])
        points = np.unique(points, axis=0)  # Remove duplicates (corners and edges of the cube)
        for p in points:
            pitch, yaw, roll = p.tolist()
            if pitch == roll == 0 and np.isclose(yaw, steer_bins).any():
                continue  # Duplicate with ground
            for boost in (0, 1):
                throttle = boost
                steer = yaw
                actions.append([throttle, steer, pitch, yaw, roll, jump, boost, handbrake])

        # Flips and jumps
        jump = handbrake = 1  # Enable handbrake for potential wavedashes
        yaw = steer = 0  # Only need roll for sideflip
        angles = [np.nan] + [v for v in flip_bins]
        for angle in angles:
            if np.isnan(angle):
                pitch = roll = 0  # Empty jump
            else:
                pitch = np.sin(angle)
                roll = np.cos(angle)
                # Project to square of diameter 2 because why not
                magnitude = max(abs(pitch), abs(roll))
                pitch /= magnitude
                roll /= magnitude
            for boost in (0, 1):
                throttle = boost
                actions.append([throttle, steer, pitch, yaw, roll, jump, boost, handbrake])
        if include_stalls:
            # Add actions for stalling
            actions.append([0, 0, 0, 1, -1, 1, 0, 1])
            actions.append([0, 0, 0, -1, 1, 1, 0, 1])

        actions = np.round(np.array(actions), 3)  # Convert to numpy and remove floating point errors
        assert len(np.unique(actions, axis=0)) == len(actions), 'Duplicate actions found'

        return actions



from typing import List, Dict, Any

from rlgym.api import SharedInfoProvider, AgentID, StateType
from rlgym.rocket_league.api import GameState


class MultiProvider(SharedInfoProvider[AgentID, GameState]):
    """
    Wrapper to use multiple SharedInfoProviders at once.
    """

    def __init__(self, *providers: SharedInfoProvider):
        self.providers = providers

    def create(self, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        for provider in self.providers:
            shared_info = provider.create(shared_info)
        return shared_info

    def set_state(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        for provider in self.providers:
            shared_info = provider.set_state(agents, initial_state, shared_info)
        return shared_info

    def step(self, agents: List[AgentID], state: StateType, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        for provider in self.providers:
            shared_info = provider.step(agents, state, shared_info)
        return shared_info



"""
    Object to contain all relevant information about the game state.
"""

import numpy as np
from typing import List, Optional

from rlgym_tools.rocket_league.v1.objects.physics_object import V1PhysicsObject
from rlgym_tools.rocket_league.v1.objects.player_data import V1PlayerData


class V1GameState(object):
    BOOST_PADS_LENGTH = 34
    BALL_STATE_LENGTH = 18
    PLAYER_CAR_STATE_LENGTH = 13
    PLAYER_TERTIARY_INFO_LENGTH = 11
    PLAYER_INFO_LENGTH = 2 + 2 * PLAYER_CAR_STATE_LENGTH + PLAYER_TERTIARY_INFO_LENGTH

    def __init__(self, state_floats: List[float] = None):
        self.game_type: int = 0
        self.blue_score: int = -1
        self.orange_score: int = -1
        self.last_touch: Optional[int] = -1

        self.players: List[V1PlayerData] = []

        self.ball: V1PhysicsObject = V1PhysicsObject()
        self.inverted_ball: V1PhysicsObject = V1PhysicsObject()

        # List of "booleans" (1 or 0)
        self.boost_pads: np.ndarray = np.zeros(V1GameState.BOOST_PADS_LENGTH, dtype=np.float32)
        self.inverted_boost_pads: np.ndarray = np.zeros_like(self.boost_pads, dtype=np.float32)

        if state_floats is not None:
            self.decode(state_floats)

    def decode(self, state_floats: List[float]):
        """
        Decode a string containing the current game state from the Bakkesmod plugin.
        :param state_floats: String containing the game state.
        """
        assert type(state_floats) == list, "UNABLE TO DECODE STATE OF TYPE {}".format(type(state_floats))
        self._decode(state_floats)

    def _decode(self, state_vals: List[float]):
        pads_len = V1GameState.BOOST_PADS_LENGTH
        p_len = V1GameState.PLAYER_INFO_LENGTH
        b_len = V1GameState.BALL_STATE_LENGTH
        start = 3

        num_ball_packets = 1
        # The state will contain the ball, the mirrored ball, every player, every player mirrored, the score for both teams, and the number of ticks since the last packet was sent.
        num_player_packets = int((len(state_vals) - num_ball_packets * b_len - start - pads_len) / p_len)

        ticks = int(state_vals[0])

        self.blue_score = int(state_vals[1])
        self.orange_score = int(state_vals[2])

        self.boost_pads[:] = state_vals[start:start + pads_len]
        self.inverted_boost_pads[:] = self.boost_pads[::-1]
        start += pads_len

        ball_data = state_vals[start:start + b_len]
        self.ball.decode_ball_data(np.asarray(ball_data))
        start += b_len // 2

        inv_ball_data = state_vals[start:start + b_len]
        self.inverted_ball.decode_ball_data(np.asarray(inv_ball_data))
        start += b_len // 2

        for i in range(num_player_packets):
            player = self._decode_player(state_vals[start:start + p_len])
            self.players.append(player)
            start += p_len

            if player.ball_touched:
                self.last_touch = player.car_id

        self.players = sorted(self.players, key=lambda p: p.car_id)  # YOU'RE WELCOME RANGLER, THIS WAS MY INNOVATION.

    def _decode_player(self, full_player_data):
        player_data = V1PlayerData()
        c_len = V1GameState.PLAYER_CAR_STATE_LENGTH
        t_len = V1GameState.PLAYER_TERTIARY_INFO_LENGTH

        start = 2

        car_data = full_player_data[start:start + c_len]
        player_data.car_data.decode_car_data(np.asarray(car_data))
        start += c_len

        inv_state_data = full_player_data[start:start + c_len]
        player_data.inverted_car_data.decode_car_data(np.asarray(inv_state_data))
        start += c_len

        tertiary_data = full_player_data[start:start + t_len]

        player_data.match_goals = int(tertiary_data[0])
        player_data.match_saves = int(tertiary_data[1])
        player_data.match_shots = int(tertiary_data[2])
        player_data.match_demolishes = int(tertiary_data[3])
        player_data.boost_pickups = int(tertiary_data[4])
        player_data.is_demoed = True if tertiary_data[5] > 0 else False
        player_data.on_ground = True if tertiary_data[6] > 0 else False
        player_data.ball_touched = True if tertiary_data[7] > 0 else False
        player_data.has_jump = True if tertiary_data[8] > 0 else False
        player_data.has_flip = True if tertiary_data[9] > 0 else False
        player_data.boost_amount = float(tertiary_data[10])
        player_data.car_id = int(full_player_data[0])
        player_data.team_num = int(full_player_data[1])

        return player_data

    def __str__(self):
        output = "{}GAME STATE OBJECT{}\n" \
                 "Game Type: {}\n" \
                 "Orange Score: {}\n" \
                 "Blue Score: {}\n" \
                 "PLAYERS: {}\n" \
                 "BALL: {}\n" \
                 "INV_BALL: {}\n" \
                 "".format("*" * 8, "*" * 8,
                           self.game_type,
                           self.orange_score,
                           self.blue_score,
                           self.players,
                           self.ball,
                           self.inverted_ball)

        return output



from dataclasses import dataclass

import numpy as np

INVERT_ACTION = np.array([1, -1, 1, -1, -1, 1, 1, 1], dtype=np.float32)  # Invert steer, yaw and roll


@dataclass(slots=True)
class Action:
    """
    Dataclass representing an action in the environment.
    RLGym uses numpy arrays, but this class can be used in place of them for better readability,
    type hints and pretty printing.
    """
    throttle: float = 0.0
    steer: float = 0.0
    pitch: float = 0.0
    yaw: float = 0.0
    roll: float = 0.0
    jump: bool = False
    boost: bool = False
    handbrake: bool = False

    @classmethod
    def from_numpy(cls, action: np.ndarray):
        return cls(*action[:5], *action[5:].astype(bool))

    def to_numpy(self):
        return np.array(
            (self.throttle, self.steer, self.pitch, self.yaw, self.roll, self.jump, self.boost, self.handbrake),
            dtype=np.float32
        )

    def mirror(self):
        return ~self

    def __invert__(self):
        # ~ operator mirrors the action
        return Action.from_numpy(self.to_numpy() * INVERT_ACTION)

    def __array__(self):
        # Make this class compatible with numpy functions
        return self.to_numpy()

    def __repr__(self):
        def format_bool(b):
            return int(b)

        def format_float(f):
            return f"{f:+.2f}"

        return (f"Action("
                f"throttle={format_float(self.throttle)}, "
                f"steer={format_float(self.steer)}, "
                f"pitch={format_float(self.pitch)}, "
                f"yaw={format_float(self.yaw)}, "
                f"roll={format_float(self.roll)}, "
                f"jump={format_bool(self.jump)}, "
                f"boost={format_bool(self.boost)}, "
                f"handbrake={format_bool(self.handbrake)}"
                f")")



from typing import Dict, Any

import numpy as np
import random

from rlgym.api import StateMutator
from rlgym.rocket_league.api import Car, GameState, PhysicsObject
from rlgym.rocket_league.common_values import BLUE_TEAM, OCTANE, ORANGE_TEAM


class VariableTeamSizeMutator(StateMutator[GameState]):

    def __init__(self, mode_weights: Dict[tuple, float]):
        self.mode_weights = mode_weights

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        assert len(state.cars) == 0, "VariableTeamSizeMutator should be applied before any cars are added to the state."
        keys, values = zip(*self.mode_weights.items())
        mode = random.choices(keys, values)[0]
        blue_size, orange_size = mode

        for idx in range(blue_size):
            car = self.make_new_car()
            car.team_num = BLUE_TEAM
            state.cars['blue-{}'.format(idx)] = car

        for idx in range(orange_size):
            car = self.make_new_car()
            car.team_num = ORANGE_TEAM
            state.cars['orange-{}'.format(idx)] = car

    def make_new_car(self) -> Car:
        """
        Users can override this to set default values.
        """
        car = Car()
        car.hitbox_type = OCTANE

        car.physics = PhysicsObject()

        car.demo_respawn_timer = 0.
        car.on_ground = True
        car.supersonic_time = 0.
        car.boost_amount = 0.
        car.boost_active_time = 0.
        car.handbrake = 0.

        car.has_jumped = False
        car.is_holding_jump = False
        car.is_jumping = False
        car.jump_time = 0.

        car.has_flipped = False
        car.has_double_jumped = False
        car.air_time_since_jump = 0.
        car.flip_time = 0.
        car.flip_torque = np.zeros(3, dtype=np.float32)

        car.is_autoflipping = False
        car.autoflip_timer = 0.
        car.autoflip_direction = 0.

        return car



from typing import List, Dict, Any

from rlgym.api import SharedInfoProvider, AgentID, StateType
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.misc.serialize import serialize


class SerializedProvider(SharedInfoProvider[AgentID, GameState]):
    def create(self, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        return shared_info

    def set_state(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        serialized_state = serialize(initial_state)
        shared_info["serialized_state"] = serialized_state
        return shared_info

    def step(self, agents: List[AgentID], state: StateType, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        serialized_state = serialize(state)
        shared_info["serialized_state"] = serialized_state
        return shared_info
